from urllib import request

import pytest
import requests
from lxml import etree
from owslib.util import ServiceException
from owslib.wcs import WebCoverageService

from datacube_ows.ows_configuration import OWSConfig, get_config
from integration_tests.utils import ODCExtent


def get_xsd(name):
    # TODO: Get XSD's for different versions
    xsd_f = request.urlopen("http://schemas.opengis.net/wcs/" + name)
    schema_doc = etree.parse(xsd_f)

    return etree.XMLSchema(schema_doc)


def check_wcs_error(
    url, expected_error_message=None, expected_status_code=400, params=None
):
    if params is None:
        params = {}
    resp = requests.get(url, params=params, timeout=10.0)
    assert resp.status_code == expected_status_code

    assert expected_error_message in resp.text
    resp_xml = etree.XML(resp.content)
    assert resp_xml is not None


def test_no_request(ows_server):
    # Make empty request to server:
    check_wcs_error(ows_server.url + "/wcs", "No operation specified", 400)


def test_no_request_wcs1(ows_server):
    # Make empty request to server:
    check_wcs_error(
        ows_server.url + "/wcs?version=1.0.0", "No operation specified", 400
    )


def test_invalid_operation(ows_server):
    # Make invalid operation request to server:
    check_wcs_error(
        ows_server.url + "/wcs?request=NoSuchOperation",
        "Unrecognised operation: NOSUCHOPERATION",
        400,
    )


def test_invalid_operation_wcs1(ows_server):
    # Make invalid operation request to server:
    check_wcs_error(
        ows_server.url + "/wcs?version=1.0.0&request=NoSuchOperation",
        "Unrecognised operation: NOSUCHOPERATION",
        400,
    )


def test_getcap_badsvc(ows_server):
    # Make bad service request to server:
    check_wcs_error(
        ows_server.url + "/wcs?request=GetCapabilities&service=NotWCS",
        "Invalid service",
        400,
    )


def test_wcs1_getcap(ows_server):
    resp = request.urlopen(
        ows_server.url + "/wcs?request=GetCapabilities&service=WCS&version=1.0.0",
        timeout=10,
    )

    # Confirm success
    assert resp.code == 200

    # Validate XML Schema
    resp_xml = etree.parse(resp.fp)
    gc_xds = get_xsd("1.0.0/wcsCapabilities.xsd")
    assert gc_xds.validate(resp_xml)


def test_wcs1_getcap_sections(ows_server):
    resp = request.urlopen(
        ows_server.url
        + "/wcs?request=GetCapabilities&service=WCS&version=1.0.0&section=/wcs_capabilities/service",
        timeout=10,
    )
    assert resp.code == 200
    resp = request.urlopen(
        ows_server.url
        + "/wcs?request=GetCapabilities&service=WCS&version=1.0.0&section=/wcs_capabilities/capability",
        timeout=10,
    )
    assert resp.code == 200
    resp = request.urlopen(
        ows_server.url
        + "/wcs?request=GetCapabilities&service=WCS&version=1.0.0&section=/wcs_capabilities/contentmetadata",
        timeout=10,
    )
    assert resp.code == 200
    check_wcs_error(
        ows_server.url + "/wcs",
        params={
            "request": "GetCapabilities",
            "service": "WCS",
            "version": "1.0.0",
            "section": "nose_bleed",
        },
        expected_error_message="Invalid section: nose_bleed",
        expected_status_code=400,
    )


def test_wcs1_server(ows_server):
    # Use owslib to confirm that we have a somewhat compliant WCS service
    wcs = WebCoverageService(url=ows_server.url + "/wcs", version="1.0.0", timeout=120)

    # Ensure that we have at least some layers available
    contents = list(wcs.contents)
    assert contents


def test_wcs1_getcov_nocov(ows_server):
    check_wcs_error(
        ows_server.url + "/wcs",
        params={
            "request": "GetCoverage",
            "service": "WCS",
            "version": "1.0.0",
        },
        expected_error_message="No coverage specified",
        expected_status_code=400,
    )


def test_wcs1_getcov_nofmt(ows_server):
    wcs = WebCoverageService(url=ows_server.url + "/wcs", version="1.0.0", timeout=120)
    contents = list(wcs.contents)
    test_layer_name = contents[0]
    check_wcs_error(
        ows_server.url + "/wcs",
        params={
            "request": "GetCoverage",
            "service": "WCS",
            "version": "1.0.0",
            "coverage": test_layer_name,
        },
        expected_error_message="No FORMAT parameter supplied",
        expected_status_code=400,
    )


def test_wcs1_getcov_bad_respcrs(ows_server):
    wcs = WebCoverageService(url=ows_server.url + "/wcs", version="1.0.0", timeout=120)
    contents = list(wcs.contents)
    test_layer_name = contents[0]
    check_wcs_error(
        ows_server.url + "/wcs",
        params={
            "request": "GetCoverage",
            "service": "WCS",
            "version": "1.0.0",
            "coverage": test_layer_name,
            "crs": "EPSG:4326",
            "response_crs": "PEGS:2346",
            "format": "GeoTIFF",
        },
        expected_error_message="PEGS:2346 is not a supported CRS",
        expected_status_code=400,
    )


def test_wcs1_getcov_nobbox(ows_server):
    wcs = WebCoverageService(url=ows_server.url + "/wcs", version="1.0.0", timeout=120)
    contents = list(wcs.contents)
    test_layer_name = contents[0]
    check_wcs_error(
        ows_server.url + "/wcs",
        params={
            "request": "GetCoverage",
            "service": "WCS",
            "version": "1.0.0",
            "coverage": test_layer_name,
            "crs": "EPSG:4326",
            "format": "GeoTIFF",
        },
        expected_error_message="No BBOX parameter supplied",
        expected_status_code=400,
    )


def test_wcs1_time_exceptions(ows_server):
    wcs = WebCoverageService(url=ows_server.url + "/wcs", version="1.0.0", timeout=120)
    contents = list(wcs.contents)
    test_layer_name = contents[0]
    cfg = get_config(refresh=True)
    layer = cfg.product_index[test_layer_name]
    extents = ODCExtent(layer).wcs1_args(
        space=ODCExtent.CENTRAL_SUBSET_FOR_TIMES, time=ODCExtent.FIRST_TWO
    )
    check_wcs_error(
        ows_server.url + "/wcs",
        params={
            "request": "GetCoverage",
            "service": "WCS",
            "version": "1.0.0",
            "coverage": test_layer_name,
            "crs": "EPSG:4326",
            "format": "GeoTIFF",
            "bbox": extents["bbox"],
            "time": "now",
        },
        expected_error_message="No valid ISO-8601 dates",
        expected_status_code=400,
    )
    check_wcs_error(
        ows_server.url + "/wcs",
        params={
            "request": "GetCoverage",
            "service": "WCS",
            "version": "1.0.0",
            "coverage": test_layer_name,
            "crs": "EPSG:4326",
            "format": "GeoTIFF",
            "bbox": extents["bbox"],
            "time": "the day after the first full moon before the Northern Solstice, in the year of our Lord Two Thousand and Twelve",
            "resx": 0.01,
            "resy": 0.01,
        },
        expected_error_message="not a valid ISO-8601 date",
        expected_status_code=400,
    )

    check_wcs_error(
        ows_server.url + "/wcs",
        params={
            "request": "GetCoverage",
            "service": "WCS",
            "version": "1.0.0",
            "coverage": test_layer_name,
            "crs": "EPSG:4326",
            "format": "GeoTIFF",
            "bbox": extents["bbox"],
            "time": "1516-04-22",
            "resx": 0.01,
            "resy": 0.01,
        },
        expected_error_message="not a valid date for coverage",
        expected_status_code=400,
    )


# Need a test database with multiple time slices to test successfully.
@pytest.mark.xfail
def test_wcs1_multi_time_exceptions(ows_server):
    wcs = WebCoverageService(url=ows_server.url + "/wcs", version="1.0.0", timeout=120)
    contents = list(wcs.contents)
    test_layer_name = contents[0]
    cfg = get_config(refresh=True)
    layer = cfg.product_index[test_layer_name]
    extents = ODCExtent(layer).wcs1_args(
        space=ODCExtent.CENTRAL_SUBSET_FOR_TIMES, time=ODCExtent.FIRST_TWO
    )
    check_wcs_error(
        ows_server.url + "/wcs",
        params={
            "request": "GetCoverage",
            "service": "WCS",
            "version": "1.0.0",
            "coverage": test_layer_name,
            "crs": "EPSG:4326",
            "format": "GeoTIFF",
            "bbox": extents["bbox"],
            "time": ",".join(extents["times"]),
            "resx": 0.01,
            "resy": 0.01,
        },
        expected_error_message="Cannot select more than one time slice with the GeoTIFF format",
        expected_status_code=400,
    )


def test_wcs1_getcov_no_meas(ows_server):
    wcs = WebCoverageService(url=ows_server.url + "/wcs", version="1.0.0", timeout=120)
    contents = list(wcs.contents)
    test_layer_name = contents[0]
    cfg = get_config(refresh=True)
    layer = cfg.product_index[test_layer_name]
    extents = ODCExtent(layer).wcs1_args(
        space=ODCExtent.CENTRAL_SUBSET_FOR_TIMES, time=ODCExtent.FIRST
    )
    check_wcs_error(
        ows_server.url + "/wcs",
        params={
            "request": "GetCoverage",
            "service": "WCS",
            "version": "1.0.0",
            "coverage": test_layer_name,
            "crs": "EPSG:4326",
            "format": "GeoTIFF",
            "bbox": extents["bbox"],
            "time": extents["times"],
            "measurements": "",
        },
        expected_error_message="No measurements supplied",
        expected_status_code=400,
    )


def test_wcs1_getcov_multi_style(ows_server):
    wcs = WebCoverageService(url=ows_server.url + "/wcs", version="1.0.0", timeout=120)
    contents = list(wcs.contents)
    test_layer_name = contents[0]
    cfg = get_config(refresh=True)
    layer = cfg.product_index[test_layer_name]
    extents = ODCExtent(layer).wcs1_args(
        space=ODCExtent.CENTRAL_SUBSET_FOR_TIMES, time=ODCExtent.FIRST
    )
    check_wcs_error(
        ows_server.url + "/wcs",
        params={
            "request": "GetCoverage",
            "service": "WCS",
            "version": "1.0.0",
            "coverage": test_layer_name,
            "crs": "EPSG:4326",
            "format": "GeoTIFF",
            "bbox": extents["bbox"],
            "time": extents["times"],
            "styles": "simple_rgb,ndvi",
        },
        expected_error_message="Multiple style parameters not supported",
        expected_status_code=400,
    )


def test_wcs1_width_height_res_exceptions(ows_server):
    wcs = WebCoverageService(url=ows_server.url + "/wcs", version="1.0.0", timeout=120)
    contents = list(wcs.contents)
    test_layer_name = contents[0]
    cfg = get_config(refresh=True)
    layer = cfg.product_index[test_layer_name]
    extents = ODCExtent(layer).wcs1_args(
        space=ODCExtent.CENTRAL_SUBSET_FOR_TIMES, time=ODCExtent.FIRST
    )
    check_wcs_error(
        ows_server.url + "/wcs",
        params={
            "request": "GetCoverage",
            "service": "WCS",
            "version": "1.0.0",
            "coverage": test_layer_name,
            "crs": "EPSG:4326",
            "format": "GeoTIFF",
            "bbox": extents["bbox"],
            "time": extents["times"],
        },
        expected_error_message="You must specify either the WIDTH/HEIGHT parameters or RESX/RESY",
        expected_status_code=400,
    )
    check_wcs_error(
        ows_server.url + "/wcs",
        params={
            "request": "GetCoverage",
            "service": "WCS",
            "version": "1.0.0",
            "coverage": test_layer_name,
            "crs": "EPSG:4326",
            "format": "GeoTIFF",
            "bbox": extents["bbox"],
            "time": extents["times"],
            "width": "200",
        },
        expected_error_message="WIDTH parameter supplied without HEIGHT parameter",
        expected_status_code=400,
    )
    check_wcs_error(
        ows_server.url + "/wcs",
        params={
            "request": "GetCoverage",
            "service": "WCS",
            "version": "1.0.0",
            "coverage": test_layer_name,
            "crs": "EPSG:4326",
            "format": "GeoTIFF",
            "bbox": extents["bbox"],
            "time": extents["times"],
            "width": "200",
            "resy": "0.0",
        },
        expected_error_message="Specify WIDTH/HEIGHT parameters OR RESX/RESY parameters - not both",
        expected_status_code=400,
    )
    check_wcs_error(
        ows_server.url + "/wcs",
        params={
            "request": "GetCoverage",
            "service": "WCS",
            "version": "1.0.0",
            "coverage": test_layer_name,
            "crs": "EPSG:4326",
            "format": "GeoTIFF",
            "bbox": extents["bbox"],
            "time": extents["times"],
            "width": "200",
            "height": "-200",
        },
        expected_error_message="HEIGHT parameter must be a positive integer",
        expected_status_code=400,
    )
    check_wcs_error(
        ows_server.url + "/wcs",
        params={
            "request": "GetCoverage",
            "service": "WCS",
            "version": "1.0.0",
            "coverage": test_layer_name,
            "crs": "EPSG:4326",
            "format": "GeoTIFF",
            "bbox": extents["bbox"],
            "time": extents["times"],
            "width": "extra thicc",
            "height": "200",
        },
        expected_error_message="WIDTH parameter must be a positive integer",
        expected_status_code=400,
    )
    check_wcs_error(
        ows_server.url + "/wcs",
        params={
            "request": "GetCoverage",
            "service": "WCS",
            "version": "1.0.0",
            "coverage": test_layer_name,
            "crs": "EPSG:4326",
            "format": "GeoTIFF",
            "bbox": extents["bbox"],
            "time": extents["times"],
            "resx": "0.01",
        },
        expected_error_message="RESX parameter supplied without RESY parameter",
        expected_status_code=400,
    )
    check_wcs_error(
        ows_server.url + "/wcs",
        params={
            "request": "GetCoverage",
            "service": "WCS",
            "version": "1.0.0",
            "coverage": test_layer_name,
            "crs": "EPSG:4326",
            "format": "GeoTIFF",
            "bbox": extents["bbox"],
            "time": extents["times"],
            "height": "300",
            "resx": "0.01",
        },
        expected_error_message="Specify WIDTH/HEIGHT parameters OR RESX/RESY parameters - not both",
        expected_status_code=400,
    )
    check_wcs_error(
        ows_server.url + "/wcs",
        params={
            "request": "GetCoverage",
            "service": "WCS",
            "version": "1.0.0",
            "coverage": test_layer_name,
            "crs": "EPSG:4326",
            "format": "GeoTIFF",
            "bbox": extents["bbox"],
            "time": extents["times"],
            "resx": "-0.01",
            "resy": "0.01",
        },
        expected_error_message="RESX parameter must be a positive number",
        expected_status_code=400,
    )
    check_wcs_error(
        ows_server.url + "/wcs",
        params={
            "request": "GetCoverage",
            "service": "WCS",
            "version": "1.0.0",
            "coverage": test_layer_name,
            "crs": "EPSG:4326",
            "format": "GeoTIFF",
            "bbox": extents["bbox"],
            "time": extents["times"],
            "resx": "0.01",
            "resy": "krunk",
        },
        expected_error_message="RESY parameter must be a positive number",
        expected_status_code=400,
    )
    check_wcs_error(
        ows_server.url + "/wcs",
        params={
            "request": "GetCoverage",
            "service": "WCS",
            "version": "1.0.0",
            "coverage": test_layer_name,
            "crs": "EPSG:4326",
            "format": "GeoTIFF",
            "bbox": extents["bbox"],
            "time": extents["times"],
            "height": "1200",
        },
        expected_error_message="HEIGHT parameter supplied without WIDTH parameter",
        expected_status_code=400,
    )
    check_wcs_error(
        ows_server.url + "/wcs",
        params={
            "request": "GetCoverage",
            "service": "WCS",
            "version": "1.0.0",
            "coverage": test_layer_name,
            "crs": "EPSG:4326",
            "format": "GeoTIFF",
            "bbox": extents["bbox"],
            "time": extents["times"],
            "resy": "1.00",
        },
        expected_error_message="RESY parameter supplied without RESX parameter",
        expected_status_code=400,
    )


def test_wcs1_style(ows_server):
    wcs = WebCoverageService(url=ows_server.url + "/wcs", version="1.0.0", timeout=120)
    contents = list(wcs.contents)
    test_layer_name = contents[0]
    cfg = get_config(refresh=True)
    layer = cfg.product_index[test_layer_name]
    extents = ODCExtent(layer).wcs1_args(
        space=ODCExtent.CENTRAL_SUBSET_FOR_TIMES, time=ODCExtent.FIRST
    )
    r = requests.get(
        ows_server.url + "/wcs",
        params={
            "request": "GetCoverage",
            "service": "WCS",
            "version": "1.0.0",
            "coverage": test_layer_name,
            "crs": "EPSG:4326",
            "format": "GeoTIFF",
            "bbox": extents["bbox"],
            "time": extents["times"],
            "styles": "simple_rgb",
            "resx": "0.01",
            "resy": "0.01",
        },
    )
    assert r.status_code == 200
    r = requests.get(
        ows_server.url + "/wcs",
        params={
            "request": "GetCoverage",
            "service": "WCS",
            "version": "1.0.0",
            "coverage": test_layer_name,
            "crs": "EPSG:4326",
            "format": "GeoTIFF",
            "bbox": extents["bbox"],
            "time": extents["times"],
            "styles": "invalid_style_returns_default_bands",
            "resx": "0.01",
            "resy": "0.01",
        },
    )
    assert r.status_code == 200


def test_wcs1_getcov_bad_meas(ows_server):
    wcs = WebCoverageService(url=ows_server.url + "/wcs", version="1.0.0", timeout=120)
    contents = list(wcs.contents)
    test_layer_name = contents[0]
    cfg = get_config(refresh=True)
    layer = cfg.product_index[test_layer_name]
    extents = ODCExtent(layer).wcs1_args(
        space=ODCExtent.CENTRAL_SUBSET_FOR_TIMES, time=ODCExtent.FIRST
    )
    check_wcs_error(
        ows_server.url + "/wcs",
        params={
            "request": "GetCoverage",
            "service": "WCS",
            "version": "1.0.0",
            "coverage": test_layer_name,
            "crs": "EPSG:4326",
            "format": "GeoTIFF",
            "bbox": extents["bbox"],
            "time": extents["times"],
            "measurements": "red,green,spam",
        },
        expected_error_message="Invalid measurement: spam",
        expected_status_code=400,
    )


def test_wcs1_getcov_badexception(ows_server):
    wcs = WebCoverageService(url=ows_server.url + "/wcs", version="1.0.0", timeout=120)
    contents = list(wcs.contents)
    test_layer_name = contents[0]
    cfg = get_config(refresh=True)
    layer = cfg.product_index[test_layer_name]
    extents = ODCExtent(layer).wcs1_args(
        space=ODCExtent.CENTRAL_SUBSET_FOR_TIMES, time=ODCExtent.FIRST
    )
    check_wcs_error(
        ows_server.url + "/wcs",
        params={
            "request": "GetCoverage",
            "service": "WCS",
            "version": "1.0.0",
            "coverage": test_layer_name,
            "crs": "EPSG:4326",
            "format": "GeoTIFF",
            "bbox": extents["bbox"],
            "time": extents["times"],
            "exceptions": "spam",
            "measurements": "red,green,blue",
        },
        expected_error_message="Unsupported exception format: spam",
        expected_status_code=400,
    )


def test_wcs1_getcov_interp(ows_server):
    wcs = WebCoverageService(url=ows_server.url + "/wcs", version="1.0.0", timeout=120)
    contents = list(wcs.contents)
    test_layer_name = contents[0]
    cfg = get_config(refresh=True)
    layer = cfg.product_index[test_layer_name]
    extents = ODCExtent(layer).wcs1_args(
        space=ODCExtent.CENTRAL_SUBSET_FOR_TIMES, time=ODCExtent.FIRST
    )
    check_wcs_error(
        ows_server.url + "/wcs",
        params={
            "request": "GetCoverage",
            "service": "WCS",
            "version": "1.0.0",
            "coverage": test_layer_name,
            "crs": "EPSG:4326",
            "format": "GeoTIFF",
            "bbox": extents["bbox"],
            "time": extents["times"],
            "exceptions": "application/vnd.ogc.se_xml",
            "measurements": "red,green,blue",
            "interpolation": "bilinear",
        },
        expected_error_message="Unsupported interpolation method: bilinear",
        expected_status_code=400,
    )


def test_wcs1_getcoverage_geotiff(ows_server):
    # Use owslib to confirm that we have a somewhat compliant WCS service
    wcs = WebCoverageService(url=ows_server.url + "/wcs", version="1.0.0", timeout=120)

    # Ensure that we have at least some layers available
    contents = list(wcs.contents)
    test_layer_name = contents[0]
    test_layer = wcs.contents[test_layer_name]

    bbox = test_layer.boundingBoxWGS84

    output = wcs.getCoverage(
        identifier=contents[0],
        format="GeoTIFF",
        bbox=pytest.helpers.disjoint_bbox(bbox),
        crs="EPSG:4326",
        width=400,
        height=300,
        timeSequence=test_layer.timepositions[
            len(test_layer.timepositions) // 2
        ].strip(),
    )

    assert output
    assert output.info()["Content-Type"] == "image/geotiff"


def test_wcs1_getcoverage_geotiff_respcrs(ows_server):
    # Use owslib to confirm that we have a somewhat compliant WCS service
    wcs = WebCoverageService(url=ows_server.url + "/wcs", version="1.0.0", timeout=120)

    # Ensure that we have at least some layers available
    contents = list(wcs.contents)
    test_layer_name = contents[0]
    test_layer = wcs.contents[test_layer_name]

    bbox = test_layer.boundingBoxWGS84

    output = wcs.getCoverage(
        identifier=contents[0],
        format="GeoTIFF",
        bbox=pytest.helpers.disjoint_bbox(bbox),
        crs="EPSG:4326",
        response_crs="EPSG:3857",
        width=400,
        height=300,
        timeSequence=test_layer.timepositions[
            len(test_layer.timepositions) // 2
        ].strip(),
    )

    assert output
    assert output.info()["Content-Type"] == "image/geotiff"


def test_wcs1_getcoverage_netcdf(ows_server):
    # Use owslib to confirm that we have a somewhat compliant WCS service
    wcs = WebCoverageService(url=ows_server.url + "/wcs", version="1.0.0", timeout=120)

    # Ensure that we have at least some layers available
    contents = list(wcs.contents)
    test_layer_name = contents[0]
    test_layer = wcs.contents[test_layer_name]

    bbox = test_layer.boundingBoxWGS84

    output = wcs.getCoverage(
        identifier=contents[0],
        format="netCDF",
        bbox=pytest.helpers.enclosed_bbox(bbox),
        crs="EPSG:4326",
        width=400,
        height=300,
    )

    assert output
    assert output.info()["Content-Type"] == "application/x-netcdf"

    output = wcs.getCoverage(
        identifier=contents[0],
        format="netCDF",
        bbox=pytest.helpers.enclosed_bbox(bbox),
        crs="I-CANT-BELIEVE-ITS-NOT-EPSG:4326",
        width=400,
        height=300,
    )

    assert output
    assert output.info()["Content-Type"] == "application/x-netcdf"


def test_extent_utils():
    OWSConfig._instance = None
    cfg = get_config(refresh=True)
    layer = None
    for lyr in cfg.product_index.values():
        if lyr.ready and not lyr.hide:
            layer = lyr
            break
    assert layer
    assert layer.ready and not layer.hide
    assert layer is not None
    ext = ODCExtent(layer)
    extent, first_times = ext.subsets(
        space=ODCExtent.FULL_LAYER_EXTENT, time=ODCExtent.FIRST
    )
    assert len(first_times) == 1
    assert extent
    assert extent == ext.full_extent
    ft_extent, last_times = ext.subsets(
        space=ODCExtent.FULL_EXTENT_FOR_TIMES, time=ODCExtent.LAST
    )
    assert len(last_times) == 1
    assert ft_extent.area < ext.full_extent.area
    assert first_times[0] < last_times[0]
    extent, times = ext.subsets(
        space=ODCExtent.CENTRAL_SUBSET_FOR_TIMES, time=ODCExtent.LAST
    )
    assert len(times) == 1
    assert extent.area < ext.full_extent.area
    assert extent.area < ft_extent.area
    assert ext.full_extent.contains(extent)
    extent, times = ext.subsets(
        space=ODCExtent.OUTSIDE_OF_FULL_EXTENT, time=ODCExtent.SECOND
    )
    assert len(times) == 1
    assert not ext.full_extent.intersects(extent)
    extent, times = ext.subsets(
        space=ODCExtent.IN_FULL_BUT_OUTSIDE_OF_TIMES, time=ODCExtent.LAST
    )
    assert not ft_extent.intersects(extent)
    assert ext.full_extent.contains(extent)


def test_wcs1_getcoverage_exceptions(ows_server):
    # Use owslib to confirm that we have a somewhat compliant WCS service
    wcs = WebCoverageService(url=ows_server.url + "/wcs", version="1.0.0", timeout=120)
    contents = list(wcs.contents)
    test_layer_name = contents[0]
    test_layer = wcs.contents[test_layer_name]

    bbox = test_layer.boundingBoxWGS84

    try:
        # test where product name is not available
        wcs.getCoverage(
            identifier="nonexistentproduct",
            format="GeoTIFF",
            bbox=pytest.helpers.disjoint_bbox(bbox),
            crs="EPSG:4326",
            width=400,
            height=300,
        )
    except ServiceException as e:
        assert "Invalid coverage:" in str(e)

    try:
        # test where  format is not supported
        wcs.getCoverage(
            identifier=contents[0],
            # format='GeoTIFF',
            bbox=pytest.helpers.disjoint_bbox(bbox),
            crs="EPSG:4326",
            width=400,
            height=300,
        )
    except ServiceException as e:
        assert "Unsupported format:" in str(e)

    try:
        # test where crs is not provided
        wcs.getCoverage(
            identifier=contents[0],
            format="GeoTIFF",
            bbox=pytest.helpers.disjoint_bbox(bbox),
            # crs='EPSG:4326',
            width=400,
            height=300,
        )
    except ServiceException as e:
        assert "No request CRS specified" in str(e)

    try:
        # test where crs is not supported
        wcs.getCoverage(
            identifier=contents[0],
            format="GeoTIFF",
            bbox=pytest.helpers.disjoint_bbox(bbox),
            crs="EPSG:432676",
            width=400,
            height=300,
        )
    except ServiceException as e:
        assert "is not a supported CRS" in str(e)

    try:
        # test where bbox is not correctly provided
        wcs.getCoverage(
            identifier=contents[0],
            format="GeoTIFF",
            # bbox=(10,40,18,45),
            crs="EPSG:4326",
            width=400,
            height=300,
        )
    except ServiceException as e:
        assert "Invalid BBOX parameter" in str(e)


def test_wcs1_describecoverage(ows_server):
    # Use owslib to confirm that we have a somewhat compliant WCS service
    wcs = WebCoverageService(url=ows_server.url + "/wcs", version="1.0.0", timeout=120)

    # Ensure that we have at least some layers available
    contents = list(wcs.contents)
    test_layer_name = contents[0]

    resp = wcs.getDescribeCoverage(test_layer_name)

    gc_xds = get_xsd("1.0.0/describeCoverage.xsd")
    assert gc_xds.validate(resp)


def test_wcs1_describecov_badcov(ows_server):
    check_wcs_error(
        ows_server.url + "/wcs",
        params={
            "request": "DescribeCoverage",
            "service": "WCS",
            "version": "1.0.0",
            "coverage": "splunge",
        },
        expected_error_message="Invalid coverage: splunge",
        expected_status_code=400,
    )


def test_wcs1_describecov_all(ows_server):
    r = requests.get(
        ows_server.url + "/wcs",
        params={
            "request": "DescribeCoverage",
            "version": "1.0.0",
        },
    )
    assert r.status_code == 200


def test_wcs20_server(ows_server):
    # Use owslib to confirm that we have a somewhat compliant WCS service
    wcs = WebCoverageService(url=ows_server.url + "/wcs", version="2.0.0", timeout=120)

    # Ensure that we have at least some layers available
    contents = list(wcs.contents)
    assert contents

    # Test DescribeCoverage
    desc_cov = wcs.getDescribeCoverage(contents[0])
    assert desc_cov


def test_wcs20_getcoverage_geotiff(ows_server):
    cfg = get_config(refresh=True)
    # Use owslib to confirm that we have a somewhat compliant WCS service
    wcs = WebCoverageService(url=ows_server.url + "/wcs", version="2.0.0", timeout=120)

    # Ensure that we have at least some layers available
    contents = list(wcs.contents)
    layer = cfg.product_index[contents[0]]
    assert layer.ready and not layer.hide
    extent = ODCExtent(layer)
    subsets = extent.wcs2_subsets(
        ODCExtent.CENTRAL_SUBSET_FOR_TIMES, ODCExtent.FIRST, "EPSG:4326"
    )
    output = wcs.getCoverage(
        identifier=[layer.name],
        format="image/geotiff",
        subsets=subsets,
        subsettingcrs="EPSG:4326",
        scalesize="x(400),y(300)",
    )
    assert output
    assert output.info()["Content-Type"] == "image/geotiff"


def test_wcs20_getcoverage_netcdf(ows_server):
    cfg = get_config(refresh=True)
    # Use owslib to confirm that we have a somewhat compliant WCS service
    wcs = WebCoverageService(url=ows_server.url + "/wcs", version="2.0.0", timeout=120)

    # Ensure that we have at least some layers available
    contents = list(wcs.contents)
    layer = cfg.product_index[contents[0]]
    extent = ODCExtent(layer)
    subsets = extent.wcs2_subsets(
        ODCExtent.CENTRAL_SUBSET_FOR_TIMES, ODCExtent.SECOND, "EPSG:4326"
    )
    output = wcs.getCoverage(
        identifier=[layer.name],
        format="application/x-netcdf",
        subsets=subsets,
        subsettingcrs="EPSG:4326",
        scalesize="x(400),y(300)",
    )

    assert output
    assert output.info()["Content-Type"] == "application/x-netcdf"


def test_wcs20_getcoverage_crs_alias(ows_server):
    cfg = get_config(refresh=True)
    # Use owslib to confirm that we have a somewhat compliant WCS service
    wcs = WebCoverageService(url=ows_server.url + "/wcs", version="2.0.0", timeout=120)

    # Ensure that we have at least some layers available
    contents = list(wcs.contents)
    layer = cfg.product_index[contents[0]]
    extent = ODCExtent(layer)
    subsets = extent.wcs2_subsets(
        ODCExtent.CENTRAL_SUBSET_FOR_TIMES, ODCExtent.SECOND_LAST, "EPSG:4326"
    )
    output = wcs.getCoverage(
        identifier=[layer.name],
        format="application/x-netcdf",
        subsets=[("x", 144, 144.3), ("y", -42.4, -42), ("time", "2019-11-05")],
        subsettingcrs="I-CANT-BELIEVE-ITS-NOT-EPSG:4326",
        scalesize="x(400),y(300)",
    )

    assert output
    assert output.info()["Content-Type"] == "application/x-netcdf"


def test_wcs20_getcoverage_multidate_geotiff(ows_server):
    cfg = get_config(refresh=True)
    # Use owslib to confirm that we have a somewhat compliant WCS service
    wcs = WebCoverageService(url=ows_server.url + "/wcs", version="2.0.0", timeout=120)

    # Ensure that we have at least some layers available
    contents = list(wcs.contents)
    layer = cfg.product_index[contents[0]]
    extent = ODCExtent(layer)
    subsets = extent.wcs2_subsets(
        ODCExtent.CENTRAL_SUBSET_FOR_TIMES, ODCExtent.FIRST_TWO, crs="EPSG:4326"
    )
    try:
        resp = wcs.getCoverage(
            identifier=[contents[0]],
            format="image/geotiff",
            subsets=subsets,
            subsettingcrs="EPSG:4326",
            scalesize="x(400),y(300)",
        )
    except ServiceException as e:
        assert "Format does not support multi-time datasets" in str(e)


def test_wcs20_getcoverage_multidate_netcdf(ows_server):
    cfg = get_config(refresh=True)
    # Use owslib to confirm that we have a somewhat compliant WCS service
    wcs = WebCoverageService(url=ows_server.url + "/wcs", version="2.0.0", timeout=120)

    # Ensure that we have at least some layers available
    contents = list(wcs.contents)
    layer = cfg.product_index[contents[0]]
    extent = ODCExtent(layer)
    subsets = extent.wcs2_subsets(
        ODCExtent.OFFSET_SUBSET_FOR_TIMES, ODCExtent.FIRST_TWO, crs="EPSG:4326"
    )
    resp = wcs.getCoverage(
        identifier=[contents[0]],
        format="application/x-netcdf",
        subsets=subsets,
        subsettingcrs="EPSG:4326",
        scalesize="x(400),y(300)",
    )


def test_wcs21_server(ows_server):
    # N.B. At time of writing owslib does not support WCS 2.1, so we have to make requests manually.
    r = requests.get(
        ows_server.url + "/wcs",
        params={"request": "GetCapabilities", "version": "2.1.0", "service": "WCS"},
    )
    assert r.status_code == 200
    cfg = get_config(refresh=True)
    layer = None
    for lyr in cfg.product_index.values():
        if lyr.ready and not lyr.hide:
            layer = lyr
            assert lyr.name in r.text
    # Ensure that we have at least some layers available
    assert layer


def test_wcs21_describecoverage(ows_server):
    cfg = get_config(refresh=True)
    layer = None
    for lyr in cfg.product_index.values():
        if lyr.ready and not lyr.hide:
            layer = lyr
            break
    assert layer
    # N.B. At time of writing owslib does not support WCS 2.1, so we have to make requests manually.
    r = requests.get(
        ows_server.url + "/wcs",
        params={
            "request": "DescribeCoverage",
            "coverageid": layer.name,
            "version": "2.1.0",
            "service": "WCS",
        },
    )
    assert r.status_code == 200
    # ETree hangs parsing schema!


#   gc_xds = get_xsd("2.1/gml/wcsDescribeCoverage.xsd")
#   assert gc_xds.validate(r.text)


def test_wcs21_getcoverage(ows_server):
    cfg = get_config(refresh=True)
    layer = None
    for lyr in cfg.product_index.values():
        if lyr.ready and not lyr.hide:
            layer = lyr
            break
    assert layer
    extent = ODCExtent(layer)
    subsets = extent.raw_wcs2_subsets(
        ODCExtent.OFFSET_SUBSET_FOR_TIMES, ODCExtent.SECOND_LAST, crs="EPSG:4326"
    )

    r = requests.get(
        ows_server.url + "/wcs",
        params={
            "request": "GetCoverage",
            "coverageid": layer.name,
            "version": "2.1.0",
            "service": "WCS",
            "format": "image/geotiff",
            "subsettingcrs": "EPSG:4326",
            "scalesize": "x(400),y(400)",
            "subset": subsets,
        },
    )
    assert r.status_code == 200


def test_wcs2_getcov_badcov(ows_server):
    check_wcs_error(
        ows_server.url + "/wcs",
        params={
            "request": "GetCoverage",
            "service": "WCS",
            "version": "2.1.0",
            "coverageid": "not_a_valid_coverage",
        },
        expected_error_message="Invalid coverage: not_a_valid_coverage",
        expected_status_code=400,
    )


def test_wcs2_getcov_unpub_subset_crs(ows_server):
    cfg = get_config(refresh=True)
    layer = None
    for lyr in cfg.product_index.values():
        if lyr.ready and not lyr.hide:
            layer = lyr
            break
    assert layer
    check_wcs_error(
        ows_server.url + "/wcs",
        params={
            "request": "GetCoverage",
            "service": "WCS",
            "version": "2.1.0",
            "coverageid": layer.name,
            "subsettingcrs": "EPSG:FAKE",
            "outputcrs": "EPSG:4326",
        },
        expected_error_message="Invalid subsettingCrs: EPSG:FAKE",
        expected_status_code=400,
    )


def test_wcs2_getcov_unpub_output_crs(ows_server):
    cfg = get_config(refresh=True)
    layer = None
    for lyr in cfg.product_index.values():
        if lyr.ready and not lyr.hide:
            layer = lyr
            break
    assert layer
    check_wcs_error(
        ows_server.url + "/wcs",
        params={
            "request": "GetCoverage",
            "service": "WCS",
            "version": "2.1.0",
            "coverageid": layer.name,
            "subsettingcrs": "EPSG:4326",
            "outputcrs": "EPSG:FAKE",
        },
        expected_error_message="Invalid outputCrs: EPSG:FAKE",
        expected_status_code=400,
    )


def test_wcs2_getcov_dup_subset_dims(ows_server):
    cfg = get_config(refresh=True)
    layer = None
    for lyr in cfg.product_index.values():
        if lyr.ready and not lyr.hide:
            layer = lyr
            break
    assert layer
    check_wcs_error(
        ows_server.url + "/wcs",
        params={
            "request": "GetCoverage",
            "service": "WCS",
            "version": "2.1.0",
            "coverageid": layer.name,
            "subsettingcrs": "EPSG:4326",
            "outputcrs": "EPSG:4326",
            "subset": (
                "x(2345234,-99999.99999)",
                "x(12315.000,-12312321)",
                "y(2345234,-99999.99999)",
                "y(12315.000,-12312321)",
            ),
        },
        expected_error_message="Duplicate dimensions:",
        expected_status_code=400,
    )


def test_wcs2_getcov_trim_time(ows_server):
    cfg = get_config(refresh=True)
    layer = None
    for lyr in cfg.product_index.values():
        if lyr.ready and not lyr.hide:
            layer = lyr
            break
    assert layer
    extent = ODCExtent(layer)
    subsets = extent.raw_wcs2_subsets(
        ODCExtent.OFFSET_SUBSET_FOR_TIMES, ODCExtent.FIRST_TWO
    )

    r = requests.get(
        ows_server.url + "/wcs",
        params={
            "request": "GetCoverage",
            "coverageid": layer.name,
            "version": "2.1.0",
            "service": "WCS",
            "format": "application/x-netcdf",
            "subsettingcrs": "EPSG:4326",
            "scalesize": "x(400),y(400)",
            "subset": subsets,
        },
    )
    assert r.status_code == 200


def test_wcs2_getcov_slice_space(ows_server):
    cfg = get_config(refresh=True)
    layer = None
    for lyr in cfg.product_index.values():
        if lyr.ready and not lyr.hide:
            layer = lyr
            break
    assert layer
    extent = ODCExtent(layer)
    subsets = list(
        extent.raw_wcs2_subsets(ODCExtent.OFFSET_SUBSET_FOR_TIMES, ODCExtent.FIRST)
    )
    subsets[0] = subsets[0].split(",")[0] + ")"

    r = requests.get(
        ows_server.url + "/wcs",
        params={
            "request": "GetCoverage",
            "coverageid": layer.name,
            "version": "2.1.0",
            "service": "WCS",
            "format": "image/geotiff",
            "subsettingcrs": "EPSG:4326",
            "scalesize": "y(400)",
            "subset": subsets,
        },
    )
    assert r.status_code == 200


def test_wcs2_getcov_invalid_space_dim(ows_server):
    cfg = get_config(refresh=True)
    layer = None
    for lyr in cfg.product_index.values():
        if lyr.ready and not lyr.hide:
            layer = lyr
            break
    assert layer
    extent = ODCExtent(layer)
    subsets = list(
        extent.raw_wcs2_subsets(ODCExtent.OFFSET_SUBSET_FOR_TIMES, ODCExtent.FIRST)
    )
    subsets[0] = "upper_fourth" + subsets[0][1:]

    check_wcs_error(
        ows_server.url + "/wcs",
        params={
            "request": "GetCoverage",
            "coverageid": layer.name,
            "version": "2.1.0",
            "service": "WCS",
            "format": "image/geotiff",
            "subsettingcrs": "EPSG:4326",
            "scalesize": "y(400)",
            "subset": subsets,
        },
        expected_error_message="Invalid subsetting axis upper_fourth",
        expected_status_code=400,
    )


def test_wcs2_getcov_duplicate_scale_dim(ows_server):
    cfg = get_config(refresh=True)
    layer = None
    for lyr in cfg.product_index.values():
        if lyr.ready and not lyr.hide:
            layer = lyr
            break
    assert layer
    extent = ODCExtent(layer)
    subsets = list(
        extent.raw_wcs2_subsets(ODCExtent.OFFSET_SUBSET_FOR_TIMES, ODCExtent.FIRST)
    )

    check_wcs_error(
        ows_server.url + "/wcs",
        params={
            "request": "GetCoverage",
            "coverageid": layer.name,
            "version": "2.1.0",
            "service": "WCS",
            "format": "image/geotiff",
            "subsettingcrs": "EPSG:4326",
            "scalesize": "x(240),y(400),y(300)",
            "subset": subsets,
        },
        expected_error_message="Duplicate scales for axis: y",
        expected_status_code=400,
    )


def test_wcs2_getcov_unscalable_dim(ows_server):
    cfg = get_config(refresh=True)
    layer = None
    for lyr in cfg.product_index.values():
        if lyr.ready and not lyr.hide:
            layer = lyr
            break
    assert layer
    extent = ODCExtent(layer)
    subsets = list(
        extent.raw_wcs2_subsets(ODCExtent.OFFSET_SUBSET_FOR_TIMES, ODCExtent.FIRST)
    )

    check_wcs_error(
        ows_server.url + "/wcs",
        params={
            "request": "GetCoverage",
            "coverageid": layer.name,
            "version": "2.1.0",
            "service": "WCS",
            "format": "image/geotiff",
            "subsettingcrs": "EPSG:4326",
            "scalesize": "x(240),y(400),time(300)",
            "subset": subsets,
        },
        expected_error_message="Cannot scale axis time",
        expected_status_code=400,
    )


def test_wcs2_getcov_bands(ows_server):
    cfg = get_config(refresh=True)
    layer = None
    for lyr in cfg.product_index.values():
        if lyr.ready and not lyr.hide:
            layer = lyr
            break
    assert layer
    extent = ODCExtent(layer)
    subsets = extent.raw_wcs2_subsets(
        ODCExtent.OFFSET_SUBSET_FOR_TIMES, ODCExtent.SECOND
    )

    r = requests.get(
        ows_server.url + "/wcs",
        params={
            "request": "GetCoverage",
            "coverageid": layer.name,
            "version": "2.1.0",
            "service": "WCS",
            "format": "image/geotiff",
            "subsettingcrs": "EPSG:4326",
            "scalesize": "x(400),y(400)",
            "rangesubset": "nir,band_3,band_6",
            "subset": subsets,
        },
    )
    assert r.status_code == 200


def test_wcs2_getcov_band_range(ows_server):
    cfg = get_config(refresh=True)
    layer = None
    for lyr in cfg.product_index.values():
        if lyr.ready and not lyr.hide:
            layer = lyr
            break
    assert layer
    extent = ODCExtent(layer)
    subsets = extent.raw_wcs2_subsets(
        ODCExtent.OFFSET_SUBSET_FOR_TIMES, ODCExtent.FIRST
    )

    r = requests.get(
        ows_server.url + "/wcs",
        params={
            "request": "GetCoverage",
            "coverageid": layer.name,
            "version": "2.1.0",
            "service": "WCS",
            "format": "image/geotiff",
            "subsettingcrs": "EPSG:4326",
            "scalesize": "x(400),y(400)",
            "rangesubset": "band_3:band_6",
            "subset": subsets,
        },
    )
    assert r.status_code == 200


def test_wcs2_getcov_bad_band(ows_server):
    cfg = get_config(refresh=True)
    layer = None
    for lyr in cfg.product_index.values():
        if lyr.ready and not lyr.hide:
            layer = lyr
            break
    assert layer
    extent = ODCExtent(layer)
    subsets = list(
        extent.raw_wcs2_subsets(ODCExtent.OFFSET_SUBSET_FOR_TIMES, ODCExtent.FIRST)
    )

    check_wcs_error(
        ows_server.url + "/wcs",
        params={
            "request": "GetCoverage",
            "coverageid": layer.name,
            "version": "2.1.0",
            "service": "WCS",
            "format": "image/geotiff",
            "subsettingcrs": "EPSG:4326",
            "scalesize": "x(240),y(400)",
            "rangesubset": "nir,green,band_6",
            "subset": subsets,
        },
        expected_error_message="No such field green",
        expected_status_code=400,
    )


def test_wcs2_getcov_bad_band_range(ows_server):
    cfg = get_config(refresh=True)
    layer = None
    for lyr in cfg.product_index.values():
        if lyr.ready and not lyr.hide:
            layer = lyr
            break
    assert layer
    extent = ODCExtent(layer)
    subsets = extent.raw_wcs2_subsets(
        ODCExtent.OFFSET_SUBSET_FOR_TIMES, ODCExtent.FIRST
    )

    check_wcs_error(
        ows_server.url + "/wcs",
        params={
            "request": "GetCoverage",
            "coverageid": layer.name,
            "version": "2.1.0",
            "service": "WCS",
            "format": "image/geotiff",
            "subsettingcrs": "EPSG:4326",
            "scalesize": "x(400),y(400)",
            "rangesubset": "green:band_6",
            "subset": subsets,
        },
        expected_error_message="No such field green",
        expected_status_code=400,
    )

    check_wcs_error(
        ows_server.url + "/wcs",
        params={
            "request": "GetCoverage",
            "coverageid": layer.name,
            "version": "2.1.0",
            "service": "WCS",
            "format": "image/geotiff",
            "subsettingcrs": "EPSG:4326",
            "scalesize": "x(400),y(400)",
            "rangesubset": "band_3:green",
            "subset": subsets,
        },
        expected_error_message="No such field green",
        expected_status_code=400,
    )


def test_wcs2_getcov_native_format(ows_server):
    cfg = get_config(refresh=True)
    layer = None
    for lyr in cfg.product_index.values():
        if lyr.ready and not lyr.hide:
            layer = lyr
            break
    assert layer
    extent = ODCExtent(layer)
    subsets = extent.raw_wcs2_subsets(
        ODCExtent.OFFSET_SUBSET_FOR_TIMES, ODCExtent.FIRST
    )

    r = requests.get(
        ows_server.url + "/wcs",
        params={
            "request": "GetCoverage",
            "coverageid": layer.name,
            "version": "2.1.0",
            "service": "WCS",
            "subsettingcrs": "EPSG:4326",
            "scalesize": "x(400),y(400)",
            "subset": subsets,
        },
    )
    assert r.status_code == 200


def test_wcs2_getcov_bad_format(ows_server):
    cfg = get_config(refresh=True)
    layer = None
    for lyr in cfg.product_index.values():
        if lyr.ready and not lyr.hide:
            layer = lyr
            break
    assert layer
    extent = ODCExtent(layer)
    subsets = extent.raw_wcs2_subsets(
        ODCExtent.OFFSET_SUBSET_FOR_TIMES, ODCExtent.FIRST
    )

    check_wcs_error(
        ows_server.url + "/wcs",
        params={
            "request": "GetCoverage",
            "coverageid": layer.name,
            "version": "2.1.0",
            "service": "WCS",
            "format": "image/reality_tv",
            "subsettingcrs": "EPSG:4326",
            "scalesize": "x(400),y(400)",
            "subset": subsets,
        },
        expected_error_message="Unsupported format: image/reality_tv",
        expected_status_code=400,
    )

    check_wcs_error(
        ows_server.url + "/wcs",
        params={
            "request": "GetCoverage",
            "coverageid": layer.name,
            "version": "2.1.0",
            "service": "WCS",
            "format": "image/geotiff",
            "subsettingcrs": "EPSG:4326",
            "scalesize": "x(400),y(400)",
            "rangesubset": "band_3:green",
            "subset": subsets,
        },
        expected_error_message="No such field green",
        expected_status_code=400,
    )


def test_wcs2_getcov_bad_multitime_format(ows_server):
    cfg = get_config(refresh=True)
    layer = None
    for lyr in cfg.product_index.values():
        if lyr.ready and not lyr.hide:
            layer = lyr
            break
    assert layer
    extent = ODCExtent(layer)
    subsets = extent.raw_wcs2_subsets(
        ODCExtent.OFFSET_SUBSET_FOR_TIMES, ODCExtent.FIRST_TWO
    )
    subsets = subsets[0:2]

    check_wcs_error(
        ows_server.url + "/wcs",
        params={
            "request": "GetCoverage",
            "coverageid": layer.name,
            "version": "2.1.0",
            "service": "WCS",
            "format": "image/geotiff",
            "subsettingcrs": "EPSG:4326",
            "scalesize": "x(400),y(400)",
            "subset": subsets,
        },
        expected_error_message="Format does not support multi-time datasets",
        expected_status_code=400,
    )



from unittest.mock import MagicMock

import pytest

from datacube_ows.config_utils import OWSConfigNotReady
from datacube_ows.ogc_utils import ConfigException
from datacube_ows.ows_configuration import BandIndex


@pytest.fixture
def minimal_prod():
    product = MagicMock()
    product.name = "foo"
    product.product_name = "foo"
    return product


def test_bidx_p_minimal(minimal_prod):
    bidx = BandIndex(minimal_prod, None)
    assert bidx.product_name == "foo"
    assert bidx.band_cfg == {}
    assert bidx._idx == {}
    assert not bidx.ready


def test_bidx_p_unready(minimal_prod):
    bidx = BandIndex(minimal_prod, {
        "foo": ["foo"]
    })
    with pytest.raises(OWSConfigNotReady) as excinfo:
        x = bidx.native_bands
    assert "native_bands" in str(excinfo.value)
    with pytest.raises(OWSConfigNotReady) as excinfo:
        x = bidx.nodata_val("foo")
    assert "_nodata_vals" in str(excinfo.value)


def test_bidx_p_duplicates(minimal_prod):
    with pytest.raises(ConfigException) as excinfo:
        bidx = BandIndex(minimal_prod, {
            "foo": ["bar"],
            "bar": ["baz"]
        })
    assert "Duplicate band name/alias" in str(excinfo.value)
    assert "bar" in str(excinfo.value)
    with pytest.raises(ConfigException) as excinfo:
        bidx = BandIndex(minimal_prod, {
            "foo": ["bar"],
            "boo": ["bar"]
        })
    assert "Duplicate band name/alias" in str(excinfo.value)
    assert "bar" in str(excinfo.value)


def test_bidx_p_band(minimal_prod):
    bidx = BandIndex(minimal_prod, {
        "foo": ["bar", "baz"],
    })
    assert bidx.band("foo") == "foo"
    assert bidx.band("bar") == "foo"
    assert bidx.band("baz") == "foo"
    with pytest.raises(ConfigException) as excinfo:
        bidx.band_label("splat")
    assert "Unknown band name/alias" in str(excinfo.value)
    assert "splat" in str(excinfo.value)


def test_bidx_p_band_labels(minimal_prod):
    bidx = BandIndex(minimal_prod, {
        "foo": ["bar", "foo", "baz"],
        "zing": ["pow", "splat"],
        "oof": [],
    })
    bls = bidx.band_labels()
    assert "bar" in bls
    assert "pow" in bls
    assert "oof" in bls
    assert len(bls) == 3


def test_bidx_p_label(minimal_prod):
    bidx = BandIndex(minimal_prod, {
        "foo": ["bar", "baz"],
    })
    assert bidx.band_label("foo") == "bar"
    assert bidx.band_label("bar") == "bar"
    assert bidx.band_label("baz") == "bar"
    with pytest.raises(ConfigException) as excinfo:
        bidx.band_label("splat")
    assert "Unknown band name/alias" in str(excinfo.value)
    assert "splat" in str(excinfo.value)


def test_bidx_makeready(minimal_prod, minimal_dc):
    bidx = BandIndex(minimal_prod, {
        "band1": [],
        "band2": ["alias2"],
        "band3": ["alias3", "band3"],
        "band4": ["band4", "alias4"]
    })
    bidx.make_ready(minimal_dc)
    assert bidx.ready
    assert bidx.band("band1") == "band1"
    assert bidx.band("alias2") == "band2"
    assert bidx.band("band3") == "band3"
    assert bidx.band("alias4") == "band4"


def test_bidx_makeready_default(minimal_prod, minimal_dc):
    bidx = BandIndex(minimal_prod, {})
    bidx.make_ready(minimal_dc)
    assert bidx.ready
    assert bidx.band("band1") == "band1"
    assert bidx.band("band2") == "band2"
    assert bidx.band("band3") == "band3"
    assert bidx.band("band4") == "band4"
    assert bidx.nodata_val("band1") == -999
    assert isinstance(bidx.nodata_val("band4"), float)


def test_bidx_makeready_invalid_band(minimal_prod, minimal_dc):
    bidx = BandIndex(minimal_prod, {
        "band1": ["band1", "valid"],
        "bandx": ["invalid"]
    })
    assert bidx.band("valid") == "band1"
    assert bidx.band("invalid") == "bandx"
    with pytest.raises(ConfigException) as excinfo:
        bidx.make_ready(minimal_dc)
    assert "Unknown band" in str(excinfo.value)
    assert "bandx" in str(excinfo.value)



import os
import sys

from datacube_ows.ows_configuration import ConfigException, read_config

src_dir = os.path.dirname(os.path.dirname(os.path.abspath(__file__)))
if src_dir not in sys.path:
    sys.path.append(src_dir)


def test_cfg_inject():
    cfg = read_config('{"test": 12345}')
    assert cfg["test"] == 12345


def test_cfg_direct(monkeypatch):
    monkeypatch.setenv("DATACUBE_OWS_CFG", "{\"test\": 12345}")
    cfg = read_config()

    assert cfg["test"] == 12345


def test_cfg_py_simple_0(monkeypatch):
    monkeypatch.setenv("DATACUBE_OWS_CFG", "tests.cfg.simple.simple")
    cfg = read_config()

    assert cfg["test"] == 123


def test_cfg_py_simple_1(monkeypatch):
    monkeypatch.setenv("DATACUBE_OWS_CFG", "tests.cfg.simple.simple1")
    cfg = read_config()

    assert cfg["test"] == 1


def test_cfg_py_nested_0(monkeypatch):
    monkeypatch.setenv("DATACUBE_OWS_CFG", "tests.cfg.nested.nested")
    cfg = read_config()

    assert cfg["test"] == 123


def test_cfg_py_nested_1(monkeypatch):
    monkeypatch.setenv("DATACUBE_OWS_CFG", "tests.cfg.nested.nested_1")
    cfg = read_config()

    assert len(cfg) == 2
    assert cfg[0]["test"] == 8888
    assert cfg[1]["test"] == 1


def test_cfg_py_nested_2(monkeypatch):
    monkeypatch.setenv("DATACUBE_OWS_CFG", "tests.cfg.nested.nested_2")
    cfg = read_config()

    assert cfg["subtest"]["test"] == 2


def test_cfg_py_nested_3(monkeypatch):
    monkeypatch.setenv("DATACUBE_OWS_CFG", "tests.cfg.nested.nested_3")
    cfg = read_config()

    assert cfg["test"] == 233
    assert len(cfg["things"]) == 3
    assert cfg["things"][0]["test"] == 2562
    assert cfg["things"][0]["thing"] is None
    assert cfg["things"][1]["test"] == 2563
    assert cfg["things"][1]["thing"]["test"] == 123
    assert cfg["things"][2]["test"] == 2564
    assert cfg["things"][2]["thing"]["test"] == 3


def test_cfg_py_nested_4(monkeypatch):
    monkeypatch.setenv("DATACUBE_OWS_CFG", "tests.cfg.nested.nested_4")
    cfg = read_config()

    assert cfg["test"] == 222
    assert len(cfg["things"]) == 3
    assert cfg["things"][0]["test"] == 2572
    assert cfg["things"][0]["thing"] is None
    assert cfg["things"][1]["test"] == 2573
    assert cfg["things"][1]["thing"]["test"] == 123
    assert cfg["things"][2]["test"] == 2574
    ncfg = cfg["things"][2]["thing"]

    assert ncfg["test"] == 233
    assert len(ncfg["things"]) == 3
    assert ncfg["things"][0]["test"] == 2562
    assert ncfg["things"][0]["thing"] is None
    assert ncfg["things"][1]["test"] == 2563
    assert ncfg["things"][1]["thing"]["test"] == 123
    assert ncfg["things"][2]["test"] == 2564
    assert ncfg["things"][2]["thing"]["test"] == 3


def test_cfg_py_infinite_1(monkeypatch):
    monkeypatch.setenv("DATACUBE_OWS_CFG", "tests.cfg.nested.infinite_1")
    try:
        cfg = read_config()
        assert False
    except ConfigException as e:
        assert str(e).startswith("Cyclic inclusion")


def test_cfg_py_infinite_2(monkeypatch):
    monkeypatch.setenv("DATACUBE_OWS_CFG", "tests.cfg.nested.infinite_2")
    try:
        cfg = read_config()
        assert False
    except ConfigException as e:
        assert str(e).startswith("Cyclic inclusion")


def test_cfg_json_simple(monkeypatch):
    monkeypatch.chdir(src_dir)
    monkeypatch.setenv("DATACUBE_OWS_CFG", "tests/cfg/nested_1.json")
    cfg = read_config()

    assert cfg["test"] == 1234


def test_cfg_json_nested_2(monkeypatch):
    monkeypatch.chdir(src_dir)
    monkeypatch.setenv("DATACUBE_OWS_CFG", "tests/cfg/nested_2.json")
    cfg = read_config()

    assert len(cfg) == 2
    assert cfg[0]["test"] == 88888
    assert cfg[1]["test"] == 1234


def validated_nested_3(cfg):
    assert cfg["test"] == 2222
    assert len(cfg["things"]) == 3
    assert cfg["things"][0]["test"] == 22562
    assert cfg["things"][0]["thing"] is None
    assert cfg["things"][1]["test"] == 22563
    assert cfg["things"][1]["thing"]["test"] == 1234
    assert cfg["things"][2]["test"] == 22564
    assert cfg["things"][2]["thing"]["test"] == 1234


def test_cfg_json_nested_3(monkeypatch):
    monkeypatch.chdir(src_dir)
    monkeypatch.setenv("DATACUBE_OWS_CFG", "tests/cfg/nested_3.json")
    cfg = read_config()
    validated_nested_3(cfg)


def test_cfg_json_nested_4(monkeypatch):
    monkeypatch.chdir(src_dir)
    monkeypatch.setenv("DATACUBE_OWS_CFG", "tests/cfg/nested_4.json")
    cfg = read_config()

    assert cfg["test"] == 3222
    assert len(cfg["things"]) == 3
    assert cfg["things"][0]["test"] == 2572
    assert cfg["things"][0]["thing"] is None
    assert cfg["things"][1]["test"] == 2573
    assert cfg["things"][1]["thing"]["test"] == 1234
    assert cfg["things"][2]["test"] == 2574
    validated_nested_3(cfg["things"][2]["thing"])


def test_cfg_json_infinite_1(monkeypatch):
    monkeypatch.chdir(src_dir)
    monkeypatch.setenv("DATACUBE_OWS_CFG", "tests/cfg/infinite_1.json")
    try:
        cfg = read_config()
        assert False
    except ConfigException as e:
        assert str(e).startswith("Cyclic inclusion")


def test_cfg_json_infinite_2(monkeypatch):
    monkeypatch.chdir(src_dir)
    monkeypatch.setenv("DATACUBE_OWS_CFG", "tests/cfg/infinite_2.json")
    try:
        cfg = read_config()
        assert False
    except ConfigException as e:
        assert str(e).startswith("Cyclic inclusion")


def test_cfg_py_mixed_1(monkeypatch):
    monkeypatch.chdir(src_dir)
    monkeypatch.setenv("DATACUBE_OWS_CFG", "tests.cfg.mixed_nested.mixed_1")
    cfg = read_config()

    assert cfg["test"] == 1234


def test_cfg_py_mixed_2(monkeypatch):
    monkeypatch.chdir(src_dir)
    monkeypatch.setenv("DATACUBE_OWS_CFG", "tests.cfg.mixed_nested.mixed_2")
    cfg = read_config()

    assert cfg["test"] == 5224
    assert cfg["subtest"]["test"] == 1234


def test_cfg_py_mixed_3(monkeypatch):
    monkeypatch.chdir(src_dir)
    monkeypatch.setenv("DATACUBE_OWS_CFG", "tests.cfg.mixed_nested.mixed_3")
    cfg = read_config()

    assert cfg["test"] == 2634
    assert cfg["subtest"]["test_py"]["test"] == 123
    assert cfg["subtest"]["test_json"]["test"] == 1234


def test_cfg_json_mixed(monkeypatch):
    monkeypatch.chdir(src_dir)
    monkeypatch.setenv("DATACUBE_OWS_CFG", "tests/cfg/mixed_nested.json")
    cfg = read_config()

    assert cfg["test"] == 9364
    assert cfg["subtest"]["test_py"]["test"] == 123
    assert cfg["subtest"]["test_json"]["test"] == 1234



from unittest.mock import MagicMock

import pytest

from datacube_ows.ogc_utils import ConfigException
from datacube_ows.ows_configuration import CacheControlRules


@pytest.fixture
def ccr_min_layer():
    lyr = MagicMock()
    lyr.name = "a_layer"
    return lyr


def test_no_rules(ccr_min_layer):
    ccr = CacheControlRules(None, ccr_min_layer, 0)
    assert ccr.rules is None
    assert not ccr.use_caching
    assert ccr.cache_headers(0) == {}
    assert ccr.cache_headers(2) == {}
    assert ccr.cache_headers(10) == {}


def test_never_cache(ccr_min_layer):
    ccr = CacheControlRules([], ccr_min_layer, 8)
    assert ccr.rules == []
    assert ccr.use_caching
    assert ccr.cache_headers(0) == {"cache-control": "no-cache"}
    assert ccr.cache_headers(2) == {"cache-control": "no-cache"}
    assert ccr.cache_headers(10) == {"cache-control": "no-cache"}


def test_complex_case(ccr_min_layer):
    ccr = CacheControlRules([
        {"min_datasets": 3, "max_age": 10000},
        {"min_datasets": 8, "max_age": 20000},
    ], ccr_min_layer, 12)
    assert ccr.use_caching
    assert ccr.cache_headers(0) == {"cache-control": "no-cache"}
    assert ccr.cache_headers(2) == {"cache-control": "no-cache"}
    assert ccr.cache_headers(3) == {"cache-control": "max-age=10000"}
    assert ccr.cache_headers(7) == {"cache-control": "max-age=10000"}
    assert ccr.cache_headers(8) == {"cache-control": "max-age=20000"}
    assert ccr.cache_headers(12) == {"cache-control": "max-age=20000"}
    assert ccr.cache_headers(13) == {"cache-control": "no-cache"}


def test_no_min_datasets_element(ccr_min_layer):
    with pytest.raises(ConfigException) as e:
        ccr = CacheControlRules([
            {"max_age": 20000},
        ], ccr_min_layer, 12)
    assert "Dataset cache rule does not contain" in str(e.value)
    assert "min_datasets" in str(e.value)
    assert "a_layer" in str(e.value)


def test_no_max_age_element(ccr_min_layer):
    with pytest.raises(ConfigException) as e:
        ccr = CacheControlRules([
            {"min_datasets": 2},
        ], ccr_min_layer, 12)
    assert "Dataset cache rule does not contain" in str(e.value)
    assert "max_age" in str(e.value)
    assert "a_layer" in str(e.value)


def test_nonint_min_ds(ccr_min_layer):
    with pytest.raises(ConfigException) as e:
        ccr = CacheControlRules([
            {"min_datasets": "2", "max_age": 2000},
        ], ccr_min_layer, 12)
    assert "Dataset cache rule" in str(e.value)
    assert "min_datasets" in str(e.value)
    assert "non-integer" in str(e.value)
    assert "a_layer" in str(e.value)


def test_nonint_max_age(ccr_min_layer):
    with pytest.raises(ConfigException) as e:
        ccr = CacheControlRules([
            {"min_datasets": 2, "max_age": 2000.5},
        ], ccr_min_layer, 12)
    assert "Dataset cache rule" in str(e.value)
    assert "max_age" in str(e.value)
    assert "non-integer" in str(e.value)
    assert "a_layer" in str(e.value)


def test_negative_min_datasets(ccr_min_layer):
    with pytest.raises(ConfigException) as e:
        ccr = CacheControlRules([
            {"min_datasets": -2, "max_age": 2000},
        ], ccr_min_layer, 12)
    assert "Invalid dataset cache rule" in str(e.value)
    assert "min_datasets" in str(e.value)
    assert "must be greater than zero" in str(e.value)
    assert "a_layer" in str(e.value)


def test_unsorted_min_datasets(ccr_min_layer):
    with pytest.raises(ConfigException) as e:
        ccr = CacheControlRules([
            {"min_datasets": 4, "max_age": 2000},
            {"min_datasets": 2, "max_age": 4000},
        ], ccr_min_layer, 12)
    assert "Dataset cache rules must be sorted" in str(e.value)
    assert "ascending" in str(e.value)
    assert "min_datasets" in str(e.value)
    assert "a_layer" in str(e.value)


def test_min_datasets_max_datasets(ccr_min_layer):
    with pytest.raises(ConfigException) as e:
        ccr = CacheControlRules([
            {"min_datasets": 2, "max_age": 2000},
            {"min_datasets": 4, "max_age": 4000},
        ], ccr_min_layer, 2)
    assert "Dataset cache rule" in str(e.value)
    assert "min_datasets" in str(e.value)
    assert "exceeds the max_datasets limit" in str(e.value)
    assert "a_layer" in str(e.value)


def test_non_negative_max_age(ccr_min_layer):
    with pytest.raises(ConfigException) as e:
        ccr = CacheControlRules([
            {"min_datasets": 2, "max_age": -2},
        ], ccr_min_layer, 12)
    assert "Dataset cache rule" in str(e.value)
    assert "max_age" in str(e.value)
    assert "must be greater than zero" in str(e.value)
    assert "a_layer" in str(e.value)


def test_unsorted_max_age(ccr_min_layer):
    with pytest.raises(ConfigException) as e:
        ccr = CacheControlRules([
            {"min_datasets": 2, "max_age": 4000},
            {"min_datasets": 4, "max_age": 2000},
        ], ccr_min_layer, 12)
    assert "dataset cache rules" in str(e.value)
    assert "must increase monotonically" in str(e.value)
    assert "max_age" in str(e.value)
    assert "a_layer" in str(e.value)




from datacube_ows.styles.api.base import (  # noqa: F401 isort:skip
    StandaloneStyle, apply_ows_style, apply_ows_style_cfg,
    generate_ows_legend_style, generate_ows_legend_style_cfg)

from datacube_ows.ogc_utils import create_geobox, xarray_image_as_png       # noqa: F401 isort:skip
from datacube_ows.band_utils import scale_data, scalable, band_modulator    # noqa: F401 isort:skip



import numpy as np
import pytest

import datacube_ows.data
from datacube_ows.data import ProductBandQuery, get_s3_browser_uris
from tests.test_styles import product_layer  # noqa: F401


@pytest.fixture
def s3_url_datasets():
    class TestDataset:
        def __init__(self, uris):
            self.uris = uris

    datasets = list()
    d1 = TestDataset([
            "s3://test-bucket/hello_world/data.yaml",
            "s3://test-bucket/hello_world/data.yaml"
    ])
    d2 = TestDataset([
            "s3://test-bucket/hello.word/foo.bar/hello.test.yaml",
            "s3://test-bucket/hello.word/foo.bar/hello-test.yaml"
    ])

    datasets.append(d1)
    datasets.append(d2)

    class DataSetMock:
        def __init__(self, datasets):
            self.datasets = datasets

            class InnerMock:
                def __init__(self, datasets):
                    self.datasets = datasets

                def item(self):
                    return self.datasets
            self.values = InnerMock(datasets)

    class PBQMock:
        def __init__(self, main):
            self.main = main

        def __hash__(self):
            return hash(self.main)

    return {
        PBQMock(True): [DataSetMock(datasets)],
        PBQMock(False): [DataSetMock(datasets)],
    }


def test_s3_browser_uris(s3_url_datasets):
    uris = get_s3_browser_uris(s3_url_datasets)

    assert "http://test-bucket.s3-website-ap-southeast-2.amazonaws.com/?prefix=hello_world" in uris
    assert "http://test-bucket.s3-website-ap-southeast-2.amazonaws.com/?prefix=hello.word/foo.bar" in uris

# TODO: read_data is now a method of the DataStacker class. This test needs a rewrite.
# @patch('xarray.Dataset')
# def test_read_data(dataset):
#
#     class fake_coords:
#         def __init__(self):
#             self.values = 1
#             self.units = "m"
#
#     class fakegeobox:
#         def __init__(self):
#             self.dimensions = ["hello"]
#             self.crs = "EPSG:3577"
#             self.coordinates = {
#                 "hello": fake_coords()
#             }
#
#     class fake_measurement:
#         def __init__(self, name, nodata, dtype):
#             self.name = name
#             self.nodata = nodata
#             self.dtype = dtype
#
#         def dataarray_attrs(self):
#             return None
#
#         def __getitem__(self, item):
#             return getattr(self, item)
#
#     class fake_dataset:
#         def __init__(self):
#             self.center_time = datetime.utcnow()
#             self.id = 1
#             self.metadata = dict()
#
#     datasets = [ fake_dataset() ]
#     measurements = [ fake_measurement("test", -1, "int16") ]
#     geobox = fakegeobox()
#     with patch('datacube.Datacube.load_data') as load_data, patch('datacube.api.query.solar_day') as solar_day:
#         datacube_wms.data.read_data(datasets, measurements, geobox)
#         assert load_data.called
#         assert solar_day.called


def test_make_derived_band_dict_nan():
    class fake_data:
        def __init__(self):
            self.nodata = np.nan

        def item(self):
            return np.nan

    class fake_dataset:
        def __getitem__(self, key):
            return fake_data()

    class fake_style:
        include_in_feature_info = True

        def __init__(self):
            self.needed_bands = ["test"]
            self.index_function = lambda x: fake_data()

    style_dict = {
        "fake": fake_style()
    }

    band_dict = datacube_ows.data._make_derived_band_dict(fake_dataset(), style_dict)
    assert band_dict["fake"] == "n/a"


def test_make_derived_band_dict_not_nan():
    class fake_data:
        def __init__(self):
            self.nodata = -6666

        def item(self):
            return 10.10

    class fake_dataset:
        def __getitem__(self, key):
            return fake_data()

    class fake_style:
        include_in_feature_info = True

        def __init__(self):
            self.needed_bands = ["test"]
            self.index_function = lambda x: fake_data()

    style_dict = {
        "fake": fake_style()
    }

    band_dict = datacube_ows.data._make_derived_band_dict(fake_dataset(), style_dict)
    assert band_dict["fake"] == 10.10


def test_make_band_dict_nan(product_layer): # noqa: F811
    class fake_data:
        def __init__(self):
            self.nodata = np.nan
            self.attrs = {}

        def item(self):
            return np.nan

    class fake_dataset:
        def __init__(self):
            self.data_vars = {
                "fake": "fake_band"
            }

        def __getitem__(self, key):
            return fake_data()

    bands = ["fake"]

    band_dict = datacube_ows.data._make_band_dict(product_layer, fake_dataset())
    assert band_dict["fake"] == "n/a"


def test_make_band_dict_float(product_layer): # noqa: F811
    import yaml
    flags_yaml = """
    flags_definition:
        category:
          bits: [0,1,2,3,4,5,6,7]
          description: Mask image as provided by JAXA - Ocean and water, lay over, shadowing, land.
          values:
            0: no_data
            50: water
            100: lay_over
            150: shadowing
            255: land
    """

    class int_data:
        def __init__(self):
            self.nodata = np.nan
            self.attrs = yaml.load(flags_yaml, yaml.Loader)

        def item(self):
            return 100

    class int_dataset:
        def __init__(self):
            self.data_vars = {
                "fake": "fake_band"
            }

        def __getitem__(self, key):
            return int_data()

    class float_data(int_data):
        def item(self):
            return 100.0

    class float_dataset(int_dataset):
        def __getitem__(self, key):
            return float_data()

    bands = ["fake"]

    band_dict = datacube_ows.data._make_band_dict(product_layer, int_dataset())
    assert isinstance(band_dict["fake"], dict)
    assert band_dict["fake"] == {
        "Mask image as provided by JAXA - Ocean and water, lay over, shadowing, land.": 'lay_over'
    }

    band_dict = datacube_ows.data._make_band_dict(product_layer, float_dataset())
    assert isinstance(band_dict["fake"], dict)
    assert band_dict["fake"] == {
        "Mask image as provided by JAXA - Ocean and water, lay over, shadowing, land.": 'lay_over'
    }


def test_pbq_ctor_simple(product_layer): # noqa: F811
    pbq = ProductBandQuery.simple_layer_query(product_layer, set(["red", "green"]))
    assert str(pbq) in (
        "Query bands {'red', 'green'} from products [FakeODCProduct(test_odc_product)]",
        "Query bands {'green', 'red'} from products [FakeODCProduct(test_odc_product)]"
    )
    pbq = ProductBandQuery.simple_layer_query(product_layer, set(["red", "green"]), resource_limited=True)
    assert str(pbq) in (
        "Query bands {'red', 'green'} from products [FakeODCProduct(test_odc_summary_product)]",
        "Query bands {'green', 'red'} from products [FakeODCProduct(test_odc_summary_product)]"
    )


def test_pbq_ctor_full(product_layer): # noqa: F811
    pbqs = ProductBandQuery.full_layer_queries(product_layer)
    assert len(pbqs) == 2
    assert "red" in str(pbqs[0])
    assert "green" in str(pbqs[0])
    assert "blue" in str(pbqs[0])
    assert "fake" in str(pbqs[0])
    assert "Query bands {" in str(pbqs[0])
    assert "} from products [FakeODCProduct(test_odc_product)]"  in str(pbqs[0])
    assert str(pbqs[1]) in (
        "Query bands ('wongle', 'pq') from products [FakeODCProduct(test_masking_product)]",
        "Query bands ('pq', 'wongle') from products [FakeODCProduct(test_masking_product)]",
    )



from datacube_ows.ows_configuration import get_config


def test_get_layers():
    layers = get_config()

    # assert layers.platforms
    # for p in layers:
    #     assert p.products
    #     for prd in p.products:
    #         assert prd.styles
    #        assert layers.product_index[prd.name] == prd
    #        assert prd.title



import numpy as np
import pytest
import xarray as xr

from datacube_ows.styles.api import (StandaloneStyle, apply_ows_style,
                                     apply_ows_style_cfg, create_geobox,
                                     generate_ows_legend_style,
                                     generate_ows_legend_style_cfg,
                                     xarray_image_as_png)
from tests.utils import coords, dim1_da, dummy_da


@pytest.fixture
def dummy_raw_data():
    output = xr.Dataset({
        "ir": dummy_da(3, "ir", coords),
        "red": dummy_da(5, "red", coords),
        "green": dummy_da(7, "green", coords),
        "blue": dummy_da(2, "blue", coords),
        "uv": dummy_da(-1, "uv", coords),
    })
    return output


@pytest.fixture
def null_mask():
    return dummy_da(True, "mask", coords, dtype=np.bool)


def test_indirect_imports():
    assert xarray_image_as_png is not None
    assert create_geobox is not None


@pytest.fixture
def simple_rgb_style_cfg():
    return {
        "name": "test_style",
        "title": "Test Style",
        "abstract": "This is a Test Style for Datacube WMS",
        "needed_bands": ["red", "green", "blue"],
        "scale_factor": 1.0,
        "components": {
            "red": {"red": 1.0},
            "green": {"green": 1.0},
            "blue": {"blue": 1.0}
        }
    }


@pytest.fixture
def simple_rgb_perband_scaling_style_cfg():
    return {
        "name": "test_style",
        "title": "Test Style",
        "abstract": "This is a Test Style for Datacube WMS",
        "needed_bands": ["red", "green", "blue"],
        "components": {
            "red": {"red": 1.0, "scale_range": [0, 200]},
            "green": {"green": 1.0, "scale_range": [0, 500]},
            "blue": {"blue": 1.0}
        },
        "scale_range": [0, 350]
    }


def test_component_style(dummy_raw_data, null_mask, simple_rgb_style_cfg):
    style = StandaloneStyle(simple_rgb_style_cfg)
    mask = style.to_mask(dummy_raw_data, null_mask)
    result = style.transform_data(dummy_raw_data, mask)
    for channel in ("red", "green", "blue"):
        assert channel in result.data_vars.keys()
    assert result["red"].values[0][0] == 5
    assert result["green"].values[0][0] == 7
    assert result["blue"].values[0][0] == 2


def test_perband_component_style(dummy_raw_data, null_mask, simple_rgb_perband_scaling_style_cfg):
    style = StandaloneStyle(simple_rgb_perband_scaling_style_cfg)
    mask = style.to_mask(dummy_raw_data, null_mask)
    result = style.transform_data(dummy_raw_data, mask)
    for channel in ("red", "green", "blue"):
        assert channel in result.data_vars.keys()


@pytest.fixture
def simple_ramp_style_cfg():
    return {
        "name": "test_style",
        "title": "Test Style",
        "abstract": "This is a Test Style for Datacube WMS",
        "index_function": {
            "function": "datacube_ows.band_utils.norm_diff",
            "mapped_bands": True,
            "kwargs": {
                "band1": "ir",
                "band2": "red"
            }
        },
        "needed_bands": ["red", "ir"],
        "color_ramp": [
            {
                "value": -0.00000001,
                "color": "#000000",
                "alpha": 0.0
            },
            {
                "value": 0.0,
                "color": "#000000",
                "alpha": 1.0
            },
            {"value": 0.2, "color": "#FF00FF"},
            {"value": 0.4, "color": "#00FF00"},
            {"value": 0.5, "color": "#FFFF00"},
            {"value": 0.6, "color": "#0000FF"},
            {"value": 0.8, "color": "#00FFFF"},
            {"value": 1.0, "color": "#FFFFFF"}
        ],
    }


@pytest.fixture
def dummy_raw_calc_data():
    dim_coords = [-2.0, -1.0, 0.0, -1.0, -2.0, -3.0]
    output = xr.Dataset({
        "ir": dim1_da("ir", [800, 100, 1000, 600, 200, 1000], dim_coords),
        "red": dim1_da("red", [200, 500, 0, 200, 200, 700], dim_coords),
        "green": dim1_da("green", [100, 500, 0, 400, 300, 200], dim_coords),
        "blue": dim1_da("blue", [200, 500, 1000, 600, 100, 700], dim_coords),
        "uv": dim1_da("uv", [400, 600, 900, 200, 400, 100], dim_coords),
        "pq": dim1_da("pq", [0b000, 0b001, 0b010, 0b011, 0b100, 0b111], dim_coords,
                      attrs={
                                "flags_definition": {
                                    "splodgy": {
                                        "bits": 2,
                                        "values": {
                                            '0': "Splodgeless",
                                            '1': "Splodgy",
                                        },
                                        "description": "All splodgy looking"
                                    },
                                    "ugly": {
                                        "bits": 1,
                                        "values": {
                                            '0': False,
                                            '1': True
                                        },
                                        "description": "Real, real ugly",
                                    },
                                    "impossible": {
                                        "bits": 0,
                                        "values": {
                                            '0': False,
                                            '1': "Woah!"
                                        },
                                        "description": "Won't happen. Can't happen. Might happen.",
                                    },
                                }
                            })
    })
    return output


def dim1_null_mask(coords):
    return dim1_da("mask", [True] * len(coords), coords)


@pytest.fixture
def raw_calc_null_mask():
    dim_coords = [-2.0, -1.0, 0.0, -1.0, -2.0, -3.0]
    return dim1_da("mask", [True] * len(dim_coords), dim_coords)


def test_ramp_style(dummy_raw_calc_data, raw_calc_null_mask, simple_ramp_style_cfg):
    style = StandaloneStyle(simple_ramp_style_cfg)
    result = apply_ows_style(style, dummy_raw_calc_data, valid_data_mask=raw_calc_null_mask)
    for channel in ("red", "green", "blue", "alpha"):
        assert channel in result.data_vars.keys()
    # point 0 800, 200 (idx=0.6)maps to blue
    assert result["alpha"].values[0] == 255
    assert result["red"].values[0] == 0
    assert result["green"].values[0] == 0
    assert result["blue"].values[0] == 255
    # point 1 100, 500 (idx<0)maps to transparent
    assert result["alpha"].values[1] == 0
    # point 2 1000,0 (idx=1.0) maps to white
    assert result["alpha"].values[2] == 255
    assert result["red"].values[2] == 255
    assert result["green"].values[2] == 255
    assert result["blue"].values[2] == 255
    # point 3 600,200 (idx=0.5) maps to yellow
    assert result["alpha"].values[3] == 255
    assert result["red"].values[3] == 255
    assert result["green"].values[3] >= 254 # Why isn't it 255?
    assert result["blue"].values[3] == 0
    # point 4 200,200 (idx=0.0) maps to black
    assert result["alpha"].values[4] == 255
    assert result["red"].values[4] == 0
    assert result["green"].values[4] == 0
    assert result["blue"].values[4] == 0
    # point 5 1000,700 (idx=0.176) maps to between black and magenta
    assert result["alpha"].values[5] == 255
    assert result["green"].values[5] == 0
    assert abs(result["red"].values[5] - result["blue"].values[5]) <= 1 # Why not exactly equal?
    assert result["red"].values[5] > 0
    assert result["red"].values[5] < 255


def test_ramp_legend_standalone(simple_ramp_style_cfg):
    style = StandaloneStyle(simple_ramp_style_cfg)
    img = generate_ows_legend_style(style, 1)
    assert img.mode == "RGBA"
    assert img.size == (400, 125)


@pytest.fixture
def rgb_style_with_masking_cfg():
    return {
        "name": "test_style",
        "title": "Test Style",
        "abstract": "This is a Test Style for Datacube WMS",
        "needed_bands": ["red", "green", "blue"],
        "scale_range": (0.0, 1000.0),
        "components": {
            "red": {"red": 1.0},
            "green": {"green": 1.0},
            "blue": {"blue": 1.0}
        },
        "pq_masks": [
            {
                "band": "pq",
                "flags": {
                    'splodgy': "Splodgeless",
                },
            },
            {
                "band": "pq",
                "flags": {
                    "ugly": True,
                    "impossible": "Woah!"
                },
                "invert": True
            },
        ]
    }


def test_component_style_with_masking(dummy_raw_calc_data, raw_calc_null_mask, rgb_style_with_masking_cfg):
    result = apply_ows_style_cfg(rgb_style_with_masking_cfg, dummy_raw_calc_data, raw_calc_null_mask)
    for channel in ("red", "green", "blue", "alpha"):
        assert channel in result.data_vars.keys()
    alphas = result["alpha"].values
    assert alphas[0] == 255
    assert alphas[1] == 255
    assert alphas[2] == 255
    assert alphas[3] == 0
    assert alphas[4] == 0
    assert alphas[5] == 0


@pytest.fixture
def dummy_col_map_data():
    dim_coords = [-2.0, -1.0, 0.0, -1.0, -2.0, -3.0]
    output = xr.Dataset({
        "pq": dim1_da("pq", [0b01000, 0b11001, 0b01010, 0b10011, 0b00100, 0b10111], dim_coords,
                      attrs={
                          "flags_definition": {
                              "joviality": {
                                  "bits": 3,
                                  "values": {
                                      '0': "Melancholic",
                                      '1': "Joyous",
                                  },
                                  "description": "All splodgy looking"
                              },
                              "flavour": {
                                  "bits": 3,
                                  "values": {
                                      '0': "Bland",
                                      '1': "Tasty",
                                  },
                                  "description": "All splodgy looking"
                              },
                              "splodgy": {
                                  "bits": 2,
                                  "values": {
                                      '0': "Splodgeless",
                                      '1': "Splodgy",
                                  },
                                  "description": "All splodgy looking"
                              },
                              "ugly": {
                                  "bits": 1,
                                  "values": {
                                      '0': False,
                                      '1': True
                                  },
                                  "description": "Real, real ugly",
                              },
                              "impossible": {
                                  "bits": 0,
                                  "values": {
                                      '0': False,
                                      '1': "Woah!"
                                  },
                                  "description": "Won't happen. Can't happen. Might happen.",
                              },
                          }
                      })
    })
    return output


@pytest.fixture
def simple_colormap_style_cfg():
    return {
        "name": "test_style",
        "title": "Test Style",
        "abstract": "This is a Test Style for Datacube WMS",
        "value_map": {
            "pq": [
                {
                    "title": "Impossibly Tasty",
                    "abstract": "Tasty AND Impossible",
                    "flags": {
                        "and": {
                            "flavour": "Tasty",
                            "impossible": "Woah!"
                        },
                    },
                    "color": "#FF0000"
                },
                {
                    "title": "Possibly Tasty",
                    "abstract": "Tasty and Possible",
                    "flags": {
                        "impossible": "Woah!"
                    },
                    "color": "#00FF00"
                },
                {
                    "title": "Ugly/Splodgy",
                    "abstract": "Ugly or splodgy",
                    "flags": {
                        "or": {
                            "ugly": True,
                            "splodgy": "Splodgy"
                        }
                    },
                    "color": "#0000FF"
                },
            ]
        }
    }


def test_colormap_style(dummy_col_map_data, raw_calc_null_mask, simple_colormap_style_cfg):
    result = apply_ows_style_cfg(simple_colormap_style_cfg, dummy_col_map_data, raw_calc_null_mask)
    for channel in ("red", "green", "blue", "alpha"):
        assert channel in result.data_vars.keys()
    # point 0 fall through - transparent
    assert result["alpha"].values[0] == 0
    # point 1 tasty & impossible: red
    assert result["alpha"].values[1] == 255
    assert result["red"].values[1] == 255
    assert result["green"].values[1] == 0
    assert result["blue"].values[1] == 0
    # point 2 splodgy or ugly: blue
    assert result["alpha"].values[2] == 255
    assert result["red"].values[2] == 0
    assert result["green"].values[2] == 0
    assert result["blue"].values[2] == 255
    # point 3 bland & impossible: green
    assert result["alpha"].values[3] == 255
    assert result["red"].values[3] == 0
    assert result["green"].values[3] == 255
    assert result["blue"].values[3] == 0
    # point 4 splodgy or ugly: blue
    assert result["alpha"].values[4] == 255
    assert result["red"].values[4] == 0
    assert result["green"].values[4] == 0
    assert result["blue"].values[4] == 255
    # point 5 bland & impossible: green
    assert result["alpha"].values[5] == 255
    assert result["red"].values[5] == 0
    assert result["green"].values[5] == 255
    assert result["blue"].values[5] == 0


@pytest.fixture
def enum_colormap_style_cfg():
    return {
        "name": "test_style",
        "title": "Test Style",
        "abstract": "This is a Test Style for Datacube WMS",
        "value_map": {
            "pq": [
                {
                    "title": "Blah",
                    "values": [8, 25],
                    "color": "#FF0000"
                },
                {
                    "title": "Rock and Roll",
                    "values": [4, 19, 25],
                    "color": "#00FF00"
                },
                {
                    "title": "",
                    "values": [23],
                    "color": "#0000FF"
                },
            ]
        }
    }


def test_enum_colormap_style(dummy_col_map_data, raw_calc_null_mask, enum_colormap_style_cfg):
    result = apply_ows_style_cfg(enum_colormap_style_cfg, dummy_col_map_data, raw_calc_null_mask)
    for channel in ("red", "green", "blue", "alpha"):
        assert channel in result.data_vars.keys()
    # point 0 (8) Blah - red
    assert result["alpha"].values[0] == 255
    assert result["red"].values[1] == 255
    assert result["green"].values[1] == 0
    assert result["blue"].values[1] == 0
    # point 1 (25) Blah - red
    assert result["alpha"].values[1] == 255
    assert result["red"].values[1] == 255
    assert result["green"].values[1] == 0
    assert result["blue"].values[1] == 0
    # point 2 (10) - fall through, transparent
    assert result["alpha"].values[2] == 0
    # point 3 (19) - rnr green
    assert result["alpha"].values[3] == 255
    assert result["red"].values[3] == 0
    assert result["green"].values[3] == 255
    assert result["blue"].values[3] == 0
    # point 4 (4): rnr green
    assert result["alpha"].values[4] == 255
    assert result["red"].values[4] == 0
    assert result["green"].values[4] == 255
    assert result["blue"].values[4] == 0
    # point 5 (23): blue
    assert result["alpha"].values[5] == 255
    assert result["red"].values[5] == 0
    assert result["green"].values[5] == 0
    assert result["blue"].values[5] == 255


def test_ramp_legend(simple_colormap_style_cfg):
    img = generate_ows_legend_style_cfg(simple_colormap_style_cfg, 1)

    assert img.mode == "RGBA"
    assert img.size == (300, 125)


def test_api_none_mask(dummy_col_map_data, raw_calc_null_mask, simple_colormap_style_cfg):
    null_mask = apply_ows_style_cfg(simple_colormap_style_cfg, dummy_col_map_data, raw_calc_null_mask)
    none_mask = apply_ows_style_cfg(simple_colormap_style_cfg, dummy_col_map_data)
    for i in range(6):
        for c in ("red", "green", "blue", "alpha"):
            assert null_mask[c].values[i] == none_mask[c].values[i]



ows_cfg = {
    "global": {
        "title": "Minimal test config",
        "allowed_urls": [],
        "info_url": "http://opendatacube.org",
        "published_CRSs": {
            "EPSG:3857": {  # Web Mercator
                "geographic": False,
                "horizontal_coord": "x",
                "vertical_coord": "y",
            },
            "EPSG:4326": {  # WGS-84
                "geographic": True,
                "vertical_coord_first": True
            },
        },
        "services": {
            "wms": True,
            "wmts": True,
            "wcs": True
        },
    },

    "wms": {},

    "wcs": {
        "formats": {
            "GeoTIFF": {
                "renderers": {
                    "1": "datacube_ows.wcs1_utils.get_tiff",
                    "2": "datacube_ows.wcs2_utils.get_tiff",
                },
                "mime": "image/geotiff",
                "extension": "tif",
                "multi-time": False
            },
            "netCDF": {
                "renderers": {
                    "1": "datacube_ows.wcs1_utils.get_netcdf",
                    "2": "datacube_ows.wcs2_utils.get_netcdf",
                },
                "mime": "application/x-netcdf",
                "extension": "nc",
                "multi-time": True,
            }
        },
        "native_format": "GeoTIFF",
    },

    "layers": [],
}







mixed_1 = {
    "include": "tests/cfg/simple.json",
    "type": "json"
}


mixed_2 = {
    "test": 5224,
    "subtest": {
        "include": "tests/cfg/simple.json",
        "type": "json"
    }
}

mixed_3 = {
    "test": 2634,
    "subtest": {
        "test_py": {
            "include": "tests.cfg.simple.simple",
            "type": "python"
        },
        "test_json": {
            "include": "tests/cfg/simple.json",
            "type": "json"
        }
    }
}



from __future__ import absolute_import, division, print_function

import logging
from functools import wraps
from time import monotonic


def log_call(func):
    @wraps(func)
    def log_wrapper(*args, **kwargs):
        _LOG = logging.getLogger()
        _LOG.debug("%s args: %s kwargs: %s", func.__name__, args, kwargs)
        return func(*args, **kwargs)
    return log_wrapper


def time_call(func):
    @wraps(func)
    def timing_wrapper(*args, **kwargs):
        start = monotonic()
        result = func(*args, **kwargs)
        stop = monotonic()
        _LOG = logging.getLogger()
        _LOG.debug("%s took: %d ms", func.__name__, int((stop - start) * 1000))
        return result
    return timing_wrapper


def group_by_statistical():
    from datacube.api.query import GroupBy

    return GroupBy(
        dimension='time',
        group_by_func=lambda ds: ds.time.begin,
        units='seconds since 1970-01-01 00:00:00',
        sort_key=lambda ds: ds.time.begin
    )


def get_sqlconn(dc):
    # pylint: disable=protected-access
    return dc.index._db._engine.connect()



import math
from unittest.mock import MagicMock, patch

import pytest

from datacube_ows.ogc_utils import ConfigException
from datacube_ows.ows_configuration import WCSFormat, parse_ows_layer


def test_native_crs_mismatch(minimal_global_cfg, minimal_layer_cfg, minimal_dc):
    minimal_global_cfg.wcs = True
    minimal_layer_cfg["wcs"] = {
        "native_crs": "EPSG:1234",
        "default_bands": ["band1", "band2", "band3"],
    }
    minimal_layer_cfg["product_name"] = "foo_nativecrs"
    lyr = parse_ows_layer(minimal_layer_cfg,
                          global_cfg=minimal_global_cfg)
    lyr.extract_bboxes = MagicMock()
    lyr.extract_bboxes.return_value = {
        "EPSG": {
            "top": 1,
            "bottom": -1,
            "left": -1,
            "right": 1,
        }
    }
    lyr.make_ready(minimal_dc)
    assert lyr.native_CRS == "EPSG:4326"


def test_native_crs_none(minimal_global_cfg, minimal_layer_cfg, minimal_dc, mock_range):
    minimal_global_cfg.wcs = True
    minimal_layer_cfg["wcs"] = {
        "default_bands": ["band1", "band2"]
    }
    minimal_layer_cfg["product_name"] = "foo_nonativecrs"
    lyr = parse_ows_layer(minimal_layer_cfg,
                          global_cfg=minimal_global_cfg)
    with patch("datacube_ows.product_ranges.get_ranges") as get_rng:
        get_rng.return_value = mock_range
        with pytest.raises(ConfigException) as excinfo:
            lyr.make_ready(minimal_dc)
    assert "a_layer" in str(excinfo.value)
    assert "No native CRS" in str(excinfo.value)


def test_native_crs_unpublished(minimal_global_cfg, minimal_layer_cfg, minimal_dc):
    minimal_global_cfg.wcs = True
    minimal_layer_cfg["wcs"] = {
        "default_bands": ["band1", "band2", "band3"],
    }
    minimal_layer_cfg["product_name"] = "foo_badnativecrs"
    lyr = parse_ows_layer(minimal_layer_cfg,
                          global_cfg=minimal_global_cfg)
    lyr.extract_bboxes = MagicMock()
    lyr.extract_bboxes.return_value = {
        "EPSG": {
            "top": 1,
            "bottom": -1,
            "left": -1,
            "right": 1,
        }
    }
    with pytest.raises(ConfigException) as excinfo:
        lyr.make_ready(minimal_dc)
    assert "EPSG:9999" in str(excinfo.value)
    assert "a_layer" in str(excinfo.value)
    assert "not in published CRSs" in str(excinfo.value)


def test_no_native_resolution(minimal_global_cfg, minimal_layer_cfg, minimal_dc, mock_range):
    minimal_global_cfg.wcs = True
    minimal_layer_cfg["wcs"] = {
        "native_crs": "EPSG:4326",
        "default_bands": ["band1", "band2", "band3"],
    }
    minimal_layer_cfg["product_name"] = "foo_nonativeres"
    lyr = parse_ows_layer(minimal_layer_cfg,
                          global_cfg=minimal_global_cfg)
    with patch("datacube_ows.product_ranges.get_ranges") as get_rng:
        get_rng.return_value = mock_range
        with pytest.raises(ConfigException) as excinfo:
            lyr.make_ready(minimal_dc)
    assert "a_layer" in str(excinfo.value)
    assert "No native resolution" in str(excinfo.value)


def test_no_native_resolution_noniter(minimal_global_cfg, minimal_layer_cfg, minimal_dc, mock_range):
    minimal_global_cfg.wcs = True
    minimal_layer_cfg["wcs"] = {
        "native_crs": "EPSG:4326",
        "default_bands": ["band1", "band2", "band3"],
        "native_resolution": 45,
    }
    minimal_layer_cfg["product_name"] = "foo_nonativeres"
    lyr = parse_ows_layer(minimal_layer_cfg,
                          global_cfg=minimal_global_cfg)
    with patch("datacube_ows.product_ranges.get_ranges") as get_rng:
        get_rng.return_value = mock_range
        with pytest.raises(ConfigException) as excinfo:
            lyr.make_ready(minimal_dc)
    assert "a_layer" in str(excinfo.value)
    assert "Invalid native resolution" in str(excinfo.value)


def test_no_native_resolution_badlen(minimal_global_cfg, minimal_layer_cfg, minimal_dc, mock_range):
    minimal_global_cfg.wcs = True
    minimal_layer_cfg["wcs"] = {
        "native_crs": "EPSG:4326",
        "default_bands": ["band1", "band2", "band3"],
        "native_resolution": [33, 45, 2234],
    }
    minimal_layer_cfg["product_name"] = "foo_nonativeres"
    lyr = parse_ows_layer(minimal_layer_cfg,
                          global_cfg=minimal_global_cfg)
    with patch("datacube_ows.product_ranges.get_ranges") as get_rng:
        get_rng.return_value = mock_range
        with pytest.raises(ConfigException) as excinfo:
            lyr.make_ready(minimal_dc)
    assert "a_layer" in str(excinfo.value)
    assert "Invalid native resolution" in str(excinfo.value)


def test_native_resolution_mismatch(minimal_global_cfg, minimal_layer_cfg, minimal_dc, mock_range):
    minimal_global_cfg.wcs = True
    minimal_layer_cfg["wcs"] = {
        "native_crs": "EPSG:4326",
        "default_bands": ["band1", "band2", "band3"],
        "native_resolution": [0.1, 0.1],
    }
    minimal_layer_cfg["product_name"] = "foo_nativeres"
    lyr = parse_ows_layer(minimal_layer_cfg,
                          global_cfg=minimal_global_cfg)
    with patch("datacube_ows.product_ranges.get_ranges") as get_rng:
        get_rng.return_value = mock_range
        lyr.make_ready(minimal_dc)
    assert not lyr.hide
    assert lyr.ready
    assert math.isclose(lyr.resolution_x, 0.001, rel_tol=1e-8)
    assert math.isclose(lyr.resolution_y, 0.001, rel_tol=1e-8)


def test_zero_grid(minimal_global_cfg, minimal_layer_cfg, minimal_dc, mock_range):
    minimal_global_cfg.wcs = True
    minimal_layer_cfg["wcs"] = {
        "native_crs": "EPSG:4326",
        "default_bands": ["band1", "band2", "band3"],
    }
    minimal_layer_cfg["product_name"] = "foo_nativeres"
    lyr = parse_ows_layer(minimal_layer_cfg,
                          global_cfg=minimal_global_cfg)
    mock_range["bboxes"]["EPSG:4326"] = {
        "top": 0.1, "bottom": 0.1,
        "left": -0.1, "right": 0.1,
    }
    with patch("datacube_ows.product_ranges.get_ranges") as get_rng:
        get_rng.return_value = mock_range
        with pytest.raises(ConfigException) as excinfo:
            lyr.make_ready(minimal_dc)
    assert not lyr.ready
    assert "Grid High y is zero" in str(excinfo.value)
    assert "a_layer" in str(excinfo.value)
    assert "EPSG:4326" in str(excinfo.value)
    lyr = parse_ows_layer(minimal_layer_cfg,
                          global_cfg=minimal_global_cfg)
    mock_range["bboxes"]["EPSG:4326"] = {
        "top": 0.1, "bottom": -0.1,
        "left": -0.1, "right": -0.1,
    }
    with patch("datacube_ows.product_ranges.get_ranges") as get_rng:
        get_rng.return_value = mock_range
        with pytest.raises(ConfigException) as excinfo:
            lyr.make_ready(minimal_dc)
    assert "Grid High x is zero" in str(excinfo.value)
    assert "a_layer" in str(excinfo.value)
    assert "EPSG:4326" in str(excinfo.value)


def test_wcs_renderer_detection():
    fmt = WCSFormat(
        "GeoTIFF",
        "image/geotiff",
        "tif",
        {
            "1": "datacube_ows.wcs1_utils.get_tiff",
            "2": "datacube_ows.wcs2_utils.get_tiff",
        },
        False
    )
    r = fmt.renderer("2.1.0")
    assert r == fmt.renderers[2]



import datacube_ows.wms_utils


def test_parse_time_delta():
    from dateutil.relativedelta import relativedelta

    tests = {
        relativedelta(hours=1): ['P0Y0M0DT1H0M0S', 'PT1H0M0S', 'PT1H', ],
        relativedelta(months=18): ['P1Y6M0DT0H0M0S', 'P1Y6M0D', 'P0Y18M0DT0H0M0S', 'P18M', ],
        relativedelta(minutes=90): ['PT1H30M', 'P0Y0M0DT1H30M0S', 'PT90M'],
    }
    for td, td_str_list in tests.items():
        for td_str in td_str_list:
            assert td == datacube_ows.wms_utils.parse_time_delta(td_str)


def test_parse_wms_time_strings():
    import datetime as dt
    tests = {
        '2018-01-10/2019-01-10': (dt.datetime(2018, 1, 10, 0, 0), dt.datetime(2019, 1, 10, 23, 23, 59, 999999)),
        '2000/P1Y': (dt.datetime(2000, 1, 1, 0, 0), dt.datetime(2000, 12, 31, 23, 59, 59, 999999)),
        '2018-01-10/P5D': (dt.datetime(2018, 1, 10, 0, 0), dt.datetime(2018, 1, 14, 23, 59, 59, 999999)),
        'P1M/2018-01-10': (dt.datetime(2017, 12, 10, 0, 0, 0, 1), dt.datetime(2018, 1, 10, 23, 23, 59, 999999)),
    }

    for value, result in tests.items():
        assert result == datacube_ows.wms_utils.parse_wms_time_strings(value.split('/'))


def test_parse_wms_time_strings_with_present():
    import datetime as dt
    start, end = datacube_ows.wms_utils.parse_wms_time_strings('2018-01-10/PRESENT'.split('/'))
    assert start == dt.datetime(2018, 1, 10, 0, 0)
    assert (dt.datetime.utcnow() - end).total_seconds() < 60



from datacube_ows.mv_index import MVSelectOpts


def test_all():
    assert MVSelectOpts.ALL.sel("Ni!!") == ["Ni!!"]


class MockSTV:
    def __init__(self, id):
        self.id = id
        self.c = self


def test_ids_datasets():
    class MockSTV:
        def __init__(self, id):
            self.id = id
            self.c = self
    stv = MockSTV(42)
    assert MVSelectOpts.IDS.sel(stv) == [42]
    assert MVSelectOpts.DATASETS.sel(stv) == [42]


def test_extent():
    sel = MVSelectOpts.EXTENT.sel(None)
    assert len(sel) == 1
    assert str(sel[0]) == "ST_AsGeoJSON(ST_Union(spatial_extent))"


def test_count():
    from sqlalchemy import text
    stv = MockSTV(id=text("foo"))
    sel = MVSelectOpts.COUNT.sel(stv)
    assert len(sel) == 1
    assert str(sel[0]) == "count(foo)"



import json
import os
from importlib import import_module
from typing import Mapping, Sequence

from datacube_ows.config_toolkit import deepinherit
from datacube_ows.ogc_utils import ConfigException, FunctionWrapper


# pylint: disable=dangerous-default-value
def cfg_expand(cfg_unexpanded, cwd=None, inclusions=[]):
    # inclusions defaulting to an empty list is dangerous, but note that it is never modified.
    # If modification of inclusions is a required, a copy (ninclusions) is made and modified instead.
    if cwd is None:
        cwd = os.getcwd()
    if isinstance(cfg_unexpanded, Mapping):
        if "include" in cfg_unexpanded:
            if cfg_unexpanded["include"] in inclusions:
                raise ConfigException("Cyclic inclusion: %s" % cfg_unexpanded["include"])
            ninclusions = inclusions.copy()
            ninclusions.append(cfg_unexpanded["include"])
            # Perform expansion
            if "type" not in cfg_unexpanded or cfg_unexpanded["type"] == "json":
                # JSON Expansion
                raw_path = cfg_unexpanded["include"]
                try:
                    # Try in actual working directory
                    json_obj = load_json_obj(raw_path)
                    abs_path = os.path.abspath(cfg_unexpanded["include"])
                    cwd = os.path.dirname(abs_path)
                # pylint: disable=broad-except
                except Exception:
                    json_obj = None
                if json_obj is None:
                    path = os.path.join(cwd, raw_path)
                    try:
                        # Try in inherited working directory
                        json_obj = load_json_obj(path)
                        abs_path = os.path.abspath(path)
                        cwd = os.path.dirname(abs_path)
                    # pylint: disable=broad-except
                    except Exception:
                        json_obj = None
                if json_obj is None:
                    raise ConfigException("Could not find json file %s" % raw_path)
                return cfg_expand(load_json_obj(abs_path), cwd=cwd, inclusions=ninclusions)
            elif cfg_unexpanded["type"] == "python":
                # Python Expansion
                return cfg_expand(import_python_obj(cfg_unexpanded["include"]), cwd=cwd, inclusions=ninclusions)
            else:
                raise ConfigException("Unsupported inclusion type: %s" % str(cfg_unexpanded["type"]))
        else:
            return {k: cfg_expand(v, cwd=cwd, inclusions=inclusions) for k, v in cfg_unexpanded.items()}
    elif isinstance(cfg_unexpanded, Sequence) and not isinstance(cfg_unexpanded, str):
        return list([cfg_expand(elem, cwd=cwd, inclusions=inclusions) for elem in cfg_unexpanded])
    else:
        return cfg_unexpanded


def load_json_obj(path):
    with open(path) as json_file:
        return json.load(json_file)


def import_python_obj(path):
    """Imports a python dictionary by fully-qualified path

    :return: a Callable object, or None
    """
    mod_name, obj_name = path.rsplit('.', 1)
    mod = import_module(mod_name)
    obj = getattr(mod, obj_name)
    return obj


class OWSConfigNotReady(ConfigException):
    pass


# Base classes for configuration objects
class OWSConfigEntry:
    # Parse and validate the json but don't access the database.
    def __init__(self, cfg, *args, **kwargs):
        self._unready_attributes = set()
        self._raw_cfg = cfg
        self.ready = False

    def declare_unready(self, name):
        if self.ready:
            raise ConfigException(f"Cannot declare {name} as unready on a ready object")
        self._unready_attributes.add(name)

    def __getattribute__(self, name):
        if name == "_unready_attributes":
            pass
        elif hasattr(self, "_unready_attributes") and name in self._unready_attributes:
            raise OWSConfigNotReady(f"The following parameters have not been initialised: {self._unready_attributes}")
        return object.__getattribute__(self, name)

    def __setattr__(self, name, val):
        if name == "_unready_attributes":
            pass
        elif hasattr(self, "_unready_attributes") and name in self._unready_attributes:
            self._unready_attributes.remove(name)
        super().__setattr__(name, val)

    # Validate against database and prepare for use.
    def make_ready(self, dc, *args, **kwargs):
        if self._unready_attributes:
            raise OWSConfigNotReady(f"The following parameters have not been initialised: {self._unready_attributes}")
        self.ready = True


class OWSEntryNotFound(ConfigException):
    pass


class OWSIndexedConfigEntry(OWSConfigEntry):
    INDEX_KEYS = []

    def __init__(self, cfg, keyvals, *args, **kwargs):
        super().__init__(cfg, *args, **kwargs)

        for k in self.INDEX_KEYS:
            if k not in keyvals:
                raise ConfigException(f"Key value {k} missing from keyvals: {keyvals!r}")
        self.keyvals = keyvals

    def lookup(self, cfg, keyvals, subs=None):
        if subs is None:
            subs = {}
        for k in self.INDEX_KEYS:
            if k not in keyvals:
                raise ConfigException(f"Key value {k} missing from keyvals: {keyvals!r}")
        return self.lookup_impl(cfg, keyvals, subs)

    @classmethod
    def lookup_impl(cls, cfg, keyvals, subs=None):
        raise NotImplementedError()


# pylint: disable=abstract-method
class OWSExtensibleConfigEntry(OWSIndexedConfigEntry):
    def __init__(self, cfg, keyvals, global_cfg, *args,
                 keyval_subs=None, keyval_defaults=None, expanded=False, **kwargs):
        if not expanded:
            cfg = self.expand_inherit(cfg, global_cfg,
                                      keyval_subs=keyval_subs, keyval_defaults=keyval_defaults)

        super().__init__(cfg, keyvals, global_cfg=global_cfg, *args, **kwargs)

    @classmethod
    def expand_inherit(cls, cfg, global_cfg, keyval_subs=None, keyval_defaults=None):
        if "inherits" in cfg:
            lookup = True
            # Precludes e.g. defaulting style lookup to current layer.
            lookup_keys = {}
            for k in cls.INDEX_KEYS:
                if k not in cfg["inherits"] and k not in keyval_defaults:
                    lookup = False
                    break
                if k in cfg["inherits"]:
                    lookup_keys[k] = cfg["inherits"][k]
                else:
                    lookup_keys[k] = keyval_defaults[k]
            if lookup:
                parent = cls.lookup_impl(global_cfg, keyvals=lookup_keys, subs=keyval_subs)
                # pylint: disable=protected-access
                parent_cfg = parent._raw_cfg
            else:
                parent_cfg = cfg["inherits"]
            cfg = deepinherit(parent_cfg, cfg)
            cfg["inheritance_expanded"] = True
        return cfg


class OWSFlagBandStandalone:
    def __init__(self, band):
        self.pq_band = band


class OWSFlagBand(OWSConfigEntry):
    def __init__(self, cfg, product_cfg, **kwargs):
        super().__init__(cfg, **kwargs)
        cfg = self._raw_cfg
        self.product = product_cfg
        pq_names = self.product.parse_pq_names(cfg)
        self.pq_names = pq_names["pq_names"]
        self.pq_low_res_names = pq_names["pq_low_res_names"]
        self.pq_band = cfg["band"]
        if "fuse_func" in cfg:
            self.pq_fuse_func = FunctionWrapper(self, cfg["fuse_func"])
        else:
            self.pq_fuse_func = None
        self.pq_ignore_time = cfg.get("ignore_time", False)
        self.ignore_info_flags = cfg.get("ignore_info_flags", [])
        self.pq_manual_merge = cfg.get("manual_merge", False)
        self.declare_unready("pq_products")
        self.declare_unready("flags_def")
        self.declare_unready("info_mask")

    # pylint: disable=attribute-defined-outside-init
    def make_ready(self, dc, *args, **kwargs):
        self.pq_products = []
        self.pq_low_res_products = []
        for pqn in self.pq_names:
            if pqn is not None:
                pq_product = dc.index.products.get_by_name(pqn)
                if pq_product is None:
                    raise ConfigException(f"Could not find flags product {pqn} for layer {self.product.name} in datacube")
                self.pq_products.append(pq_product)
        for pqn in self.pq_low_res_names:
            if pqn is not None:
                pq_product = dc.index.products.get_by_name(pqn)
                if pq_product is None:
                    raise ConfigException(f"Could not find flags low_res product {pqn} for layer {self.product.name} in datacube")
                self.pq_low_res_products.append(pq_product)

        self.info_mask = ~0
        # A (hopefully) representative product
        product = self.pq_products[0]
        meas = product.lookup_measurements([self.pq_band])
        self.flags_def = meas[self.pq_band]["flags_definition"]
        for bitname in self.ignore_info_flags:
            bit = self.flags_def[bitname]["bits"]
            if not isinstance(bit, int):
                continue
            flag = 1 << bit
            self.info_mask &= ~flag
        super().make_ready(dc)


class FlagProductBands(OWSConfigEntry):
    def __init__(self, flag_band):
        super().__init__({})
        self.bands = set()
        self.bands.add(flag_band.pq_band)
        self.flag_bands = {flag_band.pq_band: flag_band}
        self.product_names = tuple(flag_band.pq_names)
        self.ignore_time = flag_band.pq_ignore_time
        self.declare_unready("products")
        self.declare_unready("low_res_products")
        self.manual_merge = flag_band.pq_manual_merge
        self.fuse_func = flag_band.pq_fuse_func

    def products_match(self, product_names):
        return tuple(product_names) == self.product_names

    def add_flag_band(self, fb):
        self.flag_bands[fb.pq_band] = fb
        self.bands.add(fb.pq_band)
        if fb.pq_manual_merge:
            fb.pq_manual_merge = True
        if fb.pq_fuse_func and self.fuse_func and fb.pq_fuse_func != self.fuse_func:
            raise ConfigException(f"Fuse functions for flag bands in product set {self.product_names} do not match")
        if fb.pq_ignore_time != self.ignore_time:
            raise ConfigException(f"ignore_time option for flag bands in product set {self.product_names} do not match")
        elif fb.pq_fuse_func and not self.fuse_func:
            self.fuse_func = fb.pq_fuse_func
        self.declare_unready("products")
        self.declare_unready("low_res_products")

    # pylint: disable=attribute-defined-outside-init
    def make_ready(self, dc, *args, **kwargs):
        for fb in self.flag_bands.values():
            self.products = fb.pq_products
            self.low_res_products = fb.pq_low_res_products
            break
        super().make_ready(dc, *args, **kwargs)

    @classmethod
    def build_list_from_masks(cls, masks):
        flag_products = []
        for mask in masks:
            handled = False
            for fp in flag_products:
                if fp.products_match(mask.band.pq_names):
                    fp.add_flag_band(mask.band)
                    handled = True
                    break
            if not handled:
                flag_products.append(cls(mask.band))
        return flag_products

    @classmethod
    def build_list_from_flagbands(cls, flagbands):
        flag_products = []
        for fb in flagbands:
            handled = False
            for fp in flag_products:
                if fp.products_match(fb.pq_names):
                    fp.add_flag_band(fb)
                handled = True
                break
            if not handled:
                flag_products.append(cls(fb))
        return flag_products





nested = {
    "include": "tests.cfg.simple.simple",
    "type": "python",
}

nested_1 = [
    {
        "test": 8888,
    },
    {
        "include": "tests.cfg.simple.simple1",
        "type": "python"
    }
]

nested_2 = {
    "test": 3424,
    "subtest": {
        "include": "tests.cfg.simple.simple2",
        "type": "python"
    }
}

nested_3 = {
    "test": 233,
    "things": [
        {
            "test": 2562,
            "thing": None
        },
        {
            "test": 2563,
            "thing": {
                "include": "tests.cfg.simple.simple",
                "type": "python"
            }
        },
        {
            "test": 2564,
            "thing": {
                "include": "tests.cfg.simple.simple3",
                "type": "python"
            }
        },
    ]
}

nested_4 = {
    "test": 222,
    "things": [
        {
            "test": 2572,
            "thing": None
        },
        {
            "test": 2573,
            "thing": {
                "include": "tests.cfg.simple.simple",
                "type": "python"
            }
        },
        {
            "test": 2574,
            "thing": {
                "include": "tests.cfg.nested.nested_3",
                "type": "python"
            }
        },
    ]
}

infinite_1 = {
    "include": "tests.cfg.nested.infinite_1",
    "type": "python"
}


infinite_2 = {
    "test": 7777,
    "subtest": {
        "include": "tests.cfg.nested.infinite_2a",
        "type": "python"
    }
}


infinite_2a = {
    "test": 7778,
    "subtest": {
        "include": "tests.cfg.nested.infinite_2b",
        "type": "python"
    }
}


infinite_2b = {
    "test": 7779,
    "subtest": {
        "include": "tests.cfg.nested.infinite_2",
        "type": "python"
    }
}





import datetime
from unittest.mock import MagicMock, patch

import numpy as np
import pytest
from xarray import DataArray, Dataset, concat

import datacube_ows.styles
from datacube_ows.config_utils import OWSEntryNotFound
from datacube_ows.ogc_utils import ConfigException
from datacube_ows.ows_configuration import BandIndex, OWSProductLayer


@pytest.fixture
def product_layer():
    class FakeODCProduct:
        def __init__(self, name):
            self.name = name
            self.id = 7

        def __str__(self):
            return self.name

        def __repr__(self):
            return f"FakeODCProduct({self.name})"

    class FakeProductBand:
        bands = set(["pq", "wongle"])
        products = [FakeODCProduct("test_masking_product")]
        manual_merge = False
        ignore_time = False
        fuse_func = None

        def products_match(self, name):
            return False
    product_layer = OWSProductLayer.__new__(OWSProductLayer)
    product_layer._unready_attributes = []
    product_layer.global_cfg = MagicMock()
    product_layer.name = "test_product"
    product_layer.pq_band = "test_band"
    product_layer.product_names = ["test_odc_product"]
    product_layer.products = [FakeODCProduct('test_odc_product')]
    product_layer.low_res_product_names = ["test_odc_summary_product"]
    product_layer.low_res_products = [FakeODCProduct('test_odc_summary_product')]
    product_layer.always_fetch_bands = ["red", "green", "blue"]
    product_layer.band_idx = BandIndex.__new__(BandIndex)
    product_layer.band_idx._unready_attributes = []
    product_layer.band_idx.product = product_layer
    product_layer.band_idx.band_cfg = {
        "red": ["crimson", "foo", ],
        "green": [],
        "blue": ["azure", "bar"],
        "fake": []
    }
    product_layer.band_idx._idx = {
        "red": "red",
        "crimson": "red",
        "foo": "red",
        "bar": "blue",
        "green": "green",
        "blue": "blue",
        "azure": "red",
        "fake": "fake",
    }
    product_layer.global_cfg.product_index = {
        "test_product": product_layer
    }
    product_layer.data_manual_merge = False
    product_layer.fuse_func = None
    product_layer.allflag_productbands = [FakeProductBand()]
    product_layer.style_index = {}
    return product_layer


@pytest.fixture
def style_cfg_lin():
    cfg = {
        "name": "test_style",
        "title": "Test Style",
        "abstract": "This is a Test Style for Datacube WMS",
        "needed_bands": ["red", "green", "blue"],
        "scale_factor": 1.0,
        "components": {
            "red": {"red": 1.0},
            "green": {"green": 1.0},
            "blue": {"blue": 1.0}
        }
    }
    return cfg


@pytest.fixture
def style_cfg_lin_clone():
    cfg = {
        "inherits": {
            "style": "test_style",
        },
        "name": "test_style_2",
        "title": "Test Style 2",
        "scale_factor": None,
        "scale_range": [0, 12000],
    }
    return cfg


@pytest.fixture
def style_cfg_map():
    cfg = {
        "name": "test_style",
        "title": "Test Style",
        "abstract": "This is a Test Style for Datacube WMS",
        "needed_bands": ["foo"],
        "value_map": {
            "foo": [
                {
                    "title": "Invalid",
                    "abstract": "An Invalid Value",
                    "flags": {
                        "bar": True,
                        "baz": False
                    },
                    "color": "#000000"
                },
                {
                    "title": "Valid",
                    "abstract": "A Valid Value",
                    "flags": {
                        "or": {
                            "x": True,
                            "y": True
                        }
                    },
                    "color": "#FFFFFF"
                }
            ]
        }
    }
    return cfg


@pytest.fixture
def product_layer_alpha_map():
    product_layer = OWSProductLayer.__new__(OWSProductLayer)
    product_layer.global_cfg = None
    product_layer.name = "test_product"
    product_layer.pq_band = "test_band"
    product_layer.product_names = ["test_odc_product"]
    product_layer.always_fetch_bands = ["foo"]
    product_layer.band_idx = BandIndex.__new__(BandIndex)
    product_layer.band_idx.band_cfg = {
        "foo": ["foo"]
    }
    product_layer.band_idx._idx = {
        "foo": "foo"
    }
    return product_layer


@pytest.fixture
def style_cfg_map_alpha_1():
    cfg = {
        "name": "test_style",
        "title": "Test Style",
        "abstract": "This is a Test Style for Datacube WMS",
        "needed_bands": ["foo"],
        "value_map": {
            "foo": [
                {
                    "title": "Transparent",
                    "abstract": "A Transparent Value",
                    "flags": {
                        "bar": True,
                    },
                    "color": "#000000",
                    "alpha": 0.0
                }
            ]
        }
    }
    return cfg


@pytest.fixture
def style_cfg_map_alpha_2():
    cfg = {
        "name": "test_style",
        "title": "Test Style",
        "abstract": "This is a Test Style for Datacube WMS",
        "needed_bands": ["foo"],
        "value_map": {
            "foo": [
                {
                    "title": "Semi-Transparent",
                    "abstract": "A Semi-Transparent Value",
                    "flags": {
                        "bar": False,
                    },
                    "color": "#000000",
                    "alpha": 0.5
                }
            ]
        }
    }
    return cfg


@pytest.fixture
def style_cfg_map_alpha_3():
    cfg = {
        "name": "test_style",
        "title": "Test Style",
        "abstract": "This is a Test Style for Datacube WMS",
        "needed_bands": ["foo"],
        "value_map": {
            "foo": [
                {
                    "title": "Non-Transparent",
                    "abstract": "A Non-Transparent Value",
                    "flags": {
                        "bar": False,
                    },
                    "color": "#000000",
                }
            ]
        }
    }
    return cfg


@pytest.fixture
def style_cfg_ramp():
    cfg = {
        "name": "test_style",
        "title": "Test Style",
        "abstract": "This is a Test Style for Datacube WMS",
        "needed_bands": ["foo"],
        "index_function": {
            "function": "datacube_ows.band_utils.single_band",
            "mapped_bands": False,
            "kwargs": {
                "band": "foo"
            }
        },
        "color_ramp": [
            {"value": 0.0, "color": "#FFFFFF", "alpha": 0.0},
            {"value": 1.0, "color": "#000000", "alpha": 1.0}
        ]
    }
    return cfg


@pytest.fixture
def style_cfg_ramp_clone(style_cfg_ramp):
    cfg = {
        "inherits": style_cfg_ramp,
        "name": "test_style2",
        "title": "Test Style 2",
        "needed_bands": ["bar"],
    }
    return cfg


@pytest.fixture
def style_cfg_ramp_mapped():
    cfg = {
        "name": "test_style",
        "title": "Test Style",
        "abstract": "This is a Test Style for Datacube WMS",
        "needed_bands": ["foo"],
        "index_function": {
            "function": "datacube_ows.band_utils.single_band",
            "mapped_bands": True,
            "kwargs": {
                "band": "bar"
            }
        },
        "band_map": {"bar": "foo"},
        "color_ramp": [
            {"value": 0.0, "color": "#FFFFFF", "alpha": 0.0},
            {"value": 1.0, "color": "#000000", "alpha": 1.0}
        ]
    }
    return cfg


def test_correct_style_hybrid(product_layer, style_cfg_lin):
    style_cfg_lin["component_ratio"] = 1.0
    style_cfg_lin["range"] = [1, 2]
    style_cfg_lin["index_function"] = {
        "function": "datacube_ows.band_utils.constant",
        "mapped_bands": True,
        "kwargs": {
            "const": "0.1"
        }
    }
    style_def = datacube_ows.styles.StyleDef(product_layer, style_cfg_lin)

    assert isinstance(style_def, datacube_ows.styles.hybrid.HybridStyleDef)


def test_correct_style_linear(product_layer, style_cfg_lin, style_cfg_lin_clone):
    style_def = datacube_ows.styles.StyleDef(product_layer, style_cfg_lin)
    product_layer.style_index[style_def.name] = style_def


def test_style_inheritance(product_layer, style_cfg_lin, style_cfg_lin_clone):
    style_def = datacube_ows.styles.StyleDef(product_layer, style_cfg_lin)
    product_layer.style_index[style_def.name] = style_def
    style_def_clone = datacube_ows.styles.StyleDef(product_layer, style_cfg_lin_clone)
    assert isinstance(style_def_clone, datacube_ows.styles.component.ComponentStyleDef)
    assert isinstance(style_def, datacube_ows.styles.component.ComponentStyleDef)


def test_inherit_exceptions(product_layer, style_cfg_lin, style_cfg_lin_clone):
    style_def = datacube_ows.styles.StyleDef(product_layer, style_cfg_lin)
    product_layer.style_index[style_def.name] = style_def
    style_cfg_lin_clone["inherits"]["layer"] = "fake_layer"
    try:
        style_def_clone = datacube_ows.styles.StyleDef(product_layer, style_cfg_lin_clone)
        assert "Expected exception not thrown" == False
    except OWSEntryNotFound:
        pass
    try:
        style = datacube_ows.styles.StyleDef.lookup_impl(product_layer.global_cfg,
                                                         keyvals={
                                                             "style": "test_style",
                                                             "layer": "fake-layer"
                                                         })
        assert style == "Expected exception not thrown"
    except OWSEntryNotFound:
        pass
    try:
        style = datacube_ows.styles.StyleDef.lookup_impl(product_layer.global_cfg,
                                                 keyvals={
                                                     "style": "fake_style",
                                                     "layer": "test_product"
                                                 },
                                                subs={"layer": {"test_product": product_layer}})
        assert style == "Expected exception not thrown"
    except OWSEntryNotFound:
        pass


def test_style_exceptions(product_layer, style_cfg_map: dict):
    style_no_name = dict(style_cfg_map)
    style_no_name.pop('name', None)
    with pytest.raises(KeyError) as excinfo:
        style_def = datacube_ows.styles.StyleDef(product_layer, style_no_name)


def test_correct_style_map(product_layer, style_cfg_map):
    style_def = datacube_ows.styles.StyleDef(product_layer, style_cfg_map)

    assert isinstance(style_def, datacube_ows.styles.colormap.ColorMapStyleDef)


def test_alpha_style_map(
    product_layer_alpha_map,
    style_cfg_map_alpha_1,
    style_cfg_map_alpha_2,
    style_cfg_map_alpha_3):

    def fake_make_mask(data, **kwargs):
        return data


    band = np.array([True, True, True])
    timarray = [np.datetime64(datetime.date.today())]
    times = DataArray(timarray, coords=[timarray], dims=["time"], name="time")
    da = DataArray(band, name='foo')
    dst = Dataset(data_vars={'foo': da})
    ds = concat([dst], times)
    npmap = np.array([True, True, True])
    damap = DataArray(npmap)

    with patch('datacube_ows.styles.colormap.make_mask', new_callable=lambda: fake_make_mask) as fmm:
        style_def = datacube_ows.styles.StyleDef(product_layer_alpha_map, style_cfg_map_alpha_1)
        
        result = style_def.transform_data(ds, damap)
        alpha_channel = result["alpha"].values
        assert (alpha_channel == 0).all()

        style_def = datacube_ows.styles.StyleDef(product_layer_alpha_map, style_cfg_map_alpha_2)

        result = style_def.transform_data(ds, None)
        alpha_channel = result["alpha"].values
        assert (alpha_channel == 127).all()

        style_def = datacube_ows.styles.StyleDef(product_layer_alpha_map, style_cfg_map_alpha_3)

        result = style_def.transform_data(ds, damap)
        alpha_channel = result["alpha"].values
        assert (alpha_channel == 255).all()


def test_correct_style_ramp(product_layer, style_cfg_ramp):
    style_def = datacube_ows.styles.StyleDef(product_layer, style_cfg_ramp)

    assert isinstance(style_def, datacube_ows.styles.ramp.ColorRampDef)


def test_inherited_style_ramp(product_layer, style_cfg_ramp_clone):
    style_def = datacube_ows.styles.StyleDef(product_layer, style_cfg_ramp_clone)

    assert isinstance(style_def, datacube_ows.styles.ramp.ColorRampDef)


def test_bandmapped_style_ramp(product_layer, style_cfg_ramp_mapped):
    style_def = datacube_ows.styles.StyleDef(product_layer, style_cfg_ramp_mapped)

    assert isinstance(style_def, datacube_ows.styles.ramp.ColorRampDef)
    assert style_def.local_band("bar") == "red"


def test_dynamic_range_compression_scale_range(product_layer, style_cfg_lin):
    style_cfg_lin["scale_range"] = [-3000, 3000]

    style_def = datacube_ows.styles.StyleDef(product_layer, style_cfg_lin)

    assert style_def.scale_min == -3000
    assert style_def.scale_max == 3000

    band = np.zeros(3)
    band[0] = -3000
    band[1] = 0
    band[2] = 3000

    compressed = style_def.compress_band("red", band)

    assert compressed[0] == 0
    assert compressed[1] == 255 / 2
    assert compressed[2] == 255


def test_dynamic_range_compression_scale_range_clip(product_layer, style_cfg_lin):
    style_cfg_lin["scale_range"] = [-3000, 3000]

    style_def = datacube_ows.styles.StyleDef(product_layer, style_cfg_lin)

    assert style_def.scale_min == -3000
    assert style_def.scale_max == 3000

    band = np.zeros(3)
    band[0] = -3001
    band[1] = 0
    band[2] = 3001

    compressed = style_def.compress_band("red", band)

    assert compressed[0] == 0
    assert compressed[1] == 255 / 2
    assert compressed[2] == 255


def test_dynamic_range_compression_scale_factor(product_layer, style_cfg_lin):
    style_cfg_lin["scale_factor"] = 2.5

    style_def = datacube_ows.styles.StyleDef(product_layer, style_cfg_lin)

    assert style_def.scale_min == 0.0
    assert style_def.scale_max == 637.5

    band = np.zeros(3)
    band[0] = -3000
    band[1] = 0
    band[2] = 3000


@pytest.fixture
def product_layer_mask_map():
    product_layer = OWSProductLayer.__new__(OWSProductLayer)
    product_layer.global_cfg = None
    product_layer.name = "test_product"
    product_layer.pq_band = None
    product_layer.product_names = ["test_odc_product"]
    product_layer.always_fetch_bands = ["foo"]
    product_layer.band_idx = BandIndex.__new__(BandIndex)
    product_layer.band_idx.band_cfg = {
        "foo": ["foo"]
    }
    product_layer.band_idx._idx = {
        "foo": "foo"
    }
    return product_layer


@pytest.fixture
def style_cfg_map_mask():
    cfg = {
        "name": "test_style",
        "title": "Test Style",
        "abstract": "This is a Test Style for Datacube WMS",
        "needed_bands": ["foo"],
        "value_map": {
            "foo": [
                {
                    "title": "Non-Transparent",
                    "abstract": "A Non-Transparent Value",
                    "flags": {
                        "bar": 1,
                    },
                    "color": "#111111",
                    "mask": True
                },
                {
                    "title": "Non-Transparent",
                    "abstract": "A Non-Transparent Value",
                    "flags": {
                        "bar": 2,
                    },
                    "color": "#FFFFFF",
                },
                {
                    "title": "Impossible",
                    "abstract": "Will already have matched a previous rule",
                    "flags": {
                        "bar": 1,
                    },
                    "color": "#54d56f",
                }
            ]
        }
    }
    return cfg


def test_RGBAMapped_Masking(product_layer_mask_map, style_cfg_map_mask):
    def fake_make_mask(data, **kwargs):
        val = kwargs["bar"]
        return data == val


    dim = np.array([0, 1, 2, 3, 4, 5])
    band = np.array([0, 0, 1, 1, 2, 2])
    timarray = [np.datetime64(datetime.date.today())]
    times = DataArray(timarray, coords=[timarray], dims=["time"], name="time")
    da = DataArray(band, name='foo', coords={"dim": dim}, dims=["dim"])
    dst = Dataset(data_vars={'foo': da})
    ds = concat([dst], times)

    npmap = np.array([True, True, True, True, True, True])
    damap = DataArray(npmap, coords={"dim": dim}, dims=["dim"])

    with patch('datacube_ows.styles.colormap.make_mask', new_callable=lambda: fake_make_mask) as fmm:
        style_def = datacube_ows.styles.StyleDef(product_layer_mask_map, style_cfg_map_mask)
        data = style_def.transform_data(ds, damap)
        r = data["red"]
        g = data["green"]
        b = data["blue"]
        a = data["alpha"]

        assert (r.values[2] == 17)
        assert (g.values[2] == 17)
        assert (b.values[2] == 17)
        assert (a.values[2] == 0)
        assert (r.values[4] == 255)
        assert (g.values[4] == 255)
        assert (b.values[4] == 255)
        assert (a.values[4] == 255)


def test_reint():
    from datacube_ows.styles.colormap import ColorMapStyleDef

    band = np.array([0., 0., 1., 1., 2., 2.])
    da = DataArray(band, name='foo')

    assert (band.dtype.kind == "f")
    data = ColorMapStyleDef.reint(band)
    assert (data.dtype.kind == "i")

    assert (da.dtype.kind == "f")
    data = ColorMapStyleDef.reint(da)
    assert (data.dtype.kind == "i")

    data = ColorMapStyleDef.reint(data)
    assert (data.dtype.kind == "i")


def test_createcolordata():
    from colour import Color

    from datacube_ows.styles.colormap import ColorMapStyleDef

    band = np.array([0, 0, 1, 1, 2, 2])
    da = DataArray(band, name='foo')
    rgb = Color("#FFFFFF")

    data = ColorMapStyleDef.create_colordata(da, rgb, 1.0, (band >= 0))
    assert (data == 1.0).all()


def test_createcolordata_alpha():
    from colour import Color

    from datacube_ows.styles.colormap import ColorMapStyleDef

    band = np.array([0, 0, 1, 1, 2, 2])
    da = DataArray(band, name='foo')
    rgb = Color("#FFFFFF")

    data = ColorMapStyleDef.create_colordata(da, rgb, 0.0, (band >= 0))
    assert (data["alpha"] == 0).all()


def test_createcolordata_mask():
    from colour import Color

    from datacube_ows.styles.colormap import ColorMapStyleDef

    band = np.array([0, 0, 1, 1, 2, 2])
    da = DataArray(band, name='foo')
    rgb = Color("#FFFFFF")

    data = ColorMapStyleDef.create_colordata(da, rgb, 0.0, (band > 0))
    assert (np.isnan(data["red"][0:1:1])).all()
    assert (np.isfinite(data["red"][2:5:1])).all()


def test_createcolordata_remask():
    from colour import Color

    from datacube_ows.styles.colormap import ColorMapStyleDef

    band = np.array([0, 0, 1, 1, np.nan, np.nan])
    da = DataArray(band, name='foo')
    rgb = Color("#FFFFFF")

    data = ColorMapStyleDef.create_colordata(da, rgb, 0.0, np.array([True, True, True, True, True, True]))
    assert (np.isfinite(data["red"][0:3:1])).all()
    assert (np.isnan(data["red"][4:5:1])).all()


def test_scale_ramp():
    from datacube_ows.styles.ramp import scale_unscaled_ramp

    input = [
        {"value": 0.0, "color": "red", "alpha": 0.5},
        {"value": 0.5, "color": "green"},
        {"value": 1.0, "color": "blue"},
    ]
    output = scale_unscaled_ramp(-100.0, 100.0, input)
    assert output[0]["color"] == "red"
    assert output[0]["alpha"] == 0.5
    assert output[0]["value"] == -100.0
    assert output[1]["color"] == "green"
    assert output[1]["alpha"] == 1.0
    assert output[1]["value"] == 0.0
    assert output[2]["color"] == "blue"
    assert output[2]["alpha"] == 1.0
    assert output[2]["value"] == 100.0


def test_bad_mpl_ramp():
    from datacube_ows.styles.ramp import read_mpl_ramp

    with pytest.raises(ConfigException) as e:
        ramp = read_mpl_ramp("definitely_not_a_real_matplotlib_ramp_name")
    assert "Invalid Matplotlib name: " in str(e.value)
